# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides various functions to work with circles."""
## @package circles
# \ingroup draftgeoutils
# \brief Provides various functions to work with circles.

import math
import lazy_loader.lazy_loader as lz

import FreeCAD as App
import DraftVecUtils

from draftgeoutils.general import geomType, edg, vec, NORM, v1
from draftgeoutils.geometry import mirror, findDistance
from draftgeoutils.edges import findMidpoint
from draftgeoutils.intersections import findIntersection, angleBisection

# Delay import of module until first use because it is heavy
Part = lz.LazyLoader("Part", globals(), "Part")

## \addtogroup draftgeoutils
# @{


def findClosestCircle(point, circles):
    """Return the circle with closest center to a given point."""
    dist = 1000000
    closest = None
    for c in circles:
        if c.Center.sub(point).Length < dist:
            dist = c.Center.sub(point).Length
            closest = c
    return closest


def getCircleFromSpline(edge):
    """Return a circle-based edge from a bspline-based edge.

    Return None if the edge is not a BSplineCurve.
    """
    if geomType(edge) != "BSplineCurve" or len(edge.Vertexes) != 1:
        return None

    # get 2 points
    p1 = edge.Curve.value(0)
    p2 = edge.Curve.value(math.pi / 2)
    # get 2 tangents
    t1 = edge.Curve.tangent(0)[0]
    t2 = edge.Curve.tangent(math.pi / 2)[0]
    # get normal
    n = p1.cross(p2)
    if DraftVecUtils.isNull(n):
        return None

    # get rays
    r1 = DraftVecUtils.rotate(t1, math.pi / 2, n)
    r2 = DraftVecUtils.rotate(t2, math.pi / 2, n)
    # get center (intersection of rays)
    i = findIntersection(p1, p1.add(r1), p2, p2.add(r2), True, True)
    if not i:
        return None

    c = i[0]
    r = (p1.sub(c)).Length
    circle = Part.makeCircle(r, c, n)
    # print(circle.Curve)
    return circle


def circlefrom1Line2Points(edge, p1, p2):
    """Return a list of circles created from an edge and two points.

    It calculates up to 2 possible centers.
    """
    p1_p2 = edg(p1, p2)
    s = findIntersection(edge, p1_p2, True, True)
    if not s:
        return None

    s = s[0]
    v1 = p1.sub(s)
    v2 = p2.sub(s)
    projectedDist = math.sqrt(abs(v1.dot(v2)))
    edgeDir = vec(edge)
    edgeDir.normalize()

    projectedCen1 = App.Vector.add(s, App.Vector(edgeDir).multiply(projectedDist))
    projectedCen2 = App.Vector.add(s, App.Vector(edgeDir).multiply(-projectedDist))
    perpEdgeDir = edgeDir.cross(App.Vector(0, 0, 1))
    perpCen1 = App.Vector.add(projectedCen1, perpEdgeDir)
    perpCen2 = App.Vector.add(projectedCen2, perpEdgeDir)

    mid = findMidpoint(p1_p2)
    x = DraftVecUtils.crossproduct(vec(p1_p2))
    x.normalize()
    perp_mid = App.Vector.add(mid, x)
    cen1 = findIntersection(edg(projectedCen1, perpCen1), edg(mid, perp_mid), True, True)
    cen2 = findIntersection(edg(projectedCen2, perpCen2), edg(mid, perp_mid), True, True)

    circles = []
    if cen1:
        radius = DraftVecUtils.dist(projectedCen1, cen1[0])
        circles.append(Part.Circle(cen1[0], NORM, radius))

    if cen2:
        radius = DraftVecUtils.dist(projectedCen2, cen2[0])
        circles.append(Part.Circle(cen2[0], NORM, radius))

    if circles:
        return circles
    else:
        return None


def circlefrom2Lines1Point(edge1, edge2, point):
    """Return a list of circles from two edges and one point."""
    bis = angleBisection(edge1, edge2)
    if not bis:
        return None

    mirrPoint = mirror(point, bis)
    return circlefrom1Line2Points(edge1, point, mirrPoint)


def circleFrom2LinesRadius(edge1, edge2, radius):
    """Return a list of circles from two edges and one radius.

    It calculates 4 centers.
    """
    intsec = findIntersection(edge1, edge2, True, True)
    if not intsec:
        return None

    intsec = intsec[0]
    bis12 = angleBisection(edge1, edge2)
    bis21 = Part.LineSegment(
        bis12.Vertexes[0].Point, DraftVecUtils.rotate(vec(bis12), math.pi / 2.0)
    )
    ang12 = abs(DraftVecUtils.angle(vec(edge1), vec(edge2)))
    ang21 = math.pi - ang12
    dist12 = radius / math.sin(ang12 * 0.5)
    dist21 = radius / math.sin(ang21 * 0.5)

    circles = []
    cen = App.Vector.add(intsec, vec(bis12).multiply(dist12))
    circles.append(Part.Circle(cen, NORM, radius))

    cen = App.Vector.add(intsec, vec(bis12).multiply(-dist12))
    circles.append(Part.Circle(cen, NORM, radius))

    cen = App.Vector.add(intsec, vec(bis21).multiply(dist21))
    circles.append(Part.Circle(cen, NORM, radius))

    cen = App.Vector.add(intsec, vec(bis21).multiply(-dist21))
    circles.append(Part.Circle(cen, NORM, radius))

    return circles


def circleFrom3LineTangents(edge1, edge2, edge3):
    """Return a list of circles from three edges.

    It calculates up to 6 possible centers.
    """

    def rot(ed):
        geo = Part.LineSegment(v1(ed), v1(ed).add(DraftVecUtils.rotate(vec(ed), math.pi / 2)))
        return geo.toShape()

    bis12 = angleBisection(edge1, edge2)
    bis23 = angleBisection(edge2, edge3)
    bis31 = angleBisection(edge3, edge1)
    intersections = []

    intsec = findIntersection(bis12, bis23, True, True)
    if intsec:
        radius = findDistance(intsec[0], edge1).Length
        intersections.append(Part.Circle(intsec[0], NORM, radius))

    intsec = findIntersection(bis23, bis31, True, True)
    if intsec:
        radius = findDistance(intsec[0], edge1).Length
        intersections.append(Part.Circle(intsec[0], NORM, radius))

    intsec = findIntersection(bis31, bis12, True, True)
    if intsec:
        radius = findDistance(intsec[0], edge1).Length
        intersections.append(Part.Circle(intsec[0], NORM, radius))

    intsec = findIntersection(rot(bis12), rot(bis23), True, True)
    if intsec:
        radius = findDistance(intsec[0], edge1).Length
        intersections.append(Part.Circle(intsec[0], NORM, radius))

    intsec = findIntersection(rot(bis23), rot(bis31), True, True)
    if intsec:
        radius = findDistance(intsec[0], edge1).Length
        intersections.append(Part.Circle(intsec[0], NORM, radius))

    intsec = findIntersection(rot(bis31), rot(bis12), True, True)
    if intsec:
        radius = findDistance(intsec[0], edge1).Length
        intersections.append(Part.Circle(intsec[0], NORM, radius))

    circles = []
    for intsec in intersections:
        exists = False
        for cir in circles:
            if DraftVecUtils.equals(cir.Center, intsec.Center):
                exists = True
                break
        if not exists:
            circles.append(intsec)

    if circles:
        return circles
    else:
        return None


def circleFromPointLineRadius(point, edge, radius):
    """Return a list of circles from one point, one edge, and one radius.

    It calculates up to 2 possible centers.
    """
    dist = findDistance(point, edge, False)
    center1 = None
    center2 = None

    if dist.Length == 0:
        segment = vec(edge)
        perpVec = DraftVecUtils.crossproduct(segment)
        perpVec.normalize()
        normPoint_c1 = App.Vector(perpVec).multiply(radius)
        normPoint_c2 = App.Vector(perpVec).multiply(-radius)
        center1 = point.add(normPoint_c1)
        center2 = point.add(normPoint_c2)
    elif dist.Length > 2 * radius:
        return None
    elif dist.Length == 2 * radius:
        normPoint = point.add(findDistance(point, edge, False))
        dummy = (normPoint.sub(point)).multiply(0.5)
        cen = point.add(dummy)
        circ = Part.Circle(cen, NORM, radius)
        if circ:
            return [circ]
        else:
            return None
    else:
        normPoint = point.add(findDistance(point, edge, False))
        normDist = DraftVecUtils.dist(normPoint, point)
        dist = math.sqrt(radius**2 - (radius - normDist) ** 2)
        centerNormVec = DraftVecUtils.scaleTo(point.sub(normPoint), radius)
        edgeDir = edge.Vertexes[0].Point.sub(normPoint)
        edgeDir.normalize()
        center1 = centerNormVec.add(normPoint.add(App.Vector(edgeDir).multiply(dist)))
        center2 = centerNormVec.add(normPoint.add(App.Vector(edgeDir).multiply(-dist)))

    circles = []
    if center1:
        circ = Part.Circle(center1, NORM, radius)
        if circ:
            circles.append(circ)

    if center2:
        circ = Part.Circle(center2, NORM, radius)
        if circ:
            circles.append(circ)

    if circles:
        return circles
    else:
        return None


def circleFrom2PointsRadius(p1, p2, radius):
    """Return a list of circles from two points, and one radius.

    The two points must not be equal.

    It calculates up to 2 possible centers.
    """
    if DraftVecUtils.equals(p1, p2):
        return None

    p1_p2 = Part.LineSegment(p1, p2).toShape()
    dist_p1p2 = DraftVecUtils.dist(p1, p2)
    mid = findMidpoint(p1_p2)

    if dist_p1p2 == 2 * radius:
        circle = Part.Circle(mid, NORM, radius)
        if circle:
            return [circle]
        else:
            return None

    _dir = vec(p1_p2)
    _dir.normalize()
    perpDir = _dir.cross(App.Vector(0, 0, 1))
    perpDir.normalize()
    dist = math.sqrt(radius**2 - (dist_p1p2 / 2.0) ** 2)
    cen1 = App.Vector.add(mid, App.Vector(perpDir).multiply(dist))
    cen2 = App.Vector.add(mid, App.Vector(perpDir).multiply(-dist))

    circles = []
    if cen1:
        circles.append(Part.Circle(cen1, NORM, radius))
    if cen2:
        circles.append(Part.Circle(cen2, NORM, radius))

    if circles:
        return circles
    else:
        return None


def findHomotheticCenterOfCircles(circle1, circle2):
    """Calculate the homothetic centers from two circles.

    Return None if the objects are not circles, or if they are concentric.

    http://en.wikipedia.org/wiki/Homothetic_center
    http://mathworld.wolfram.com/HomotheticCenter.html
    """
    if geomType(circle1) == "Circle" and geomType(circle2) == "Circle":
        print("debug: findHomotheticCenterOfCircles bad parameters!")
        return None

    if DraftVecUtils.equals(circle1.Curve.Center, circle2.Curve.Center):
        return None

    cen1_cen2 = Part.LineSegment(circle1.Curve.Center, circle2.Curve.Center).toShape()
    cenDir = vec(cen1_cen2)
    cenDir.normalize()

    # Get the perpendicular vector.
    perpCenDir = cenDir.cross(App.Vector(0, 0, 1))
    perpCenDir.normalize()

    # Get point on first circle
    p1 = App.Vector.add(circle1.Curve.Center, App.Vector(perpCenDir).multiply(circle1.Curve.Radius))

    centers = []
    # Calculate inner homothetic center
    # Get point on second circle
    p2_inner = App.Vector.add(
        circle1.Curve.Center, App.Vector(perpCenDir).multiply(-circle1.Curve.Radius)
    )
    hCenterInner = DraftVecUtils.intersect(
        circle1.Curve.Center, circle2.Curve.Center, p1, p2_inner, True, True
    )
    if hCenterInner:
        centers.append(hCenterInner)

    # Calculate outer homothetic center; it only exists if the circles
    # have different radii
    if circle1.Curve.Radius != circle2.Curve.Radius:
        # Get point on second circle
        p2_outer = App.Vector.add(
            circle1.Curve.Center, App.Vector(perpCenDir).multiply(circle1.Curve.Radius)
        )
        hCenterOuter = DraftVecUtils.intersect(
            circle1.Curve.Center, circle2.Curve.Center, p1, p2_outer, True, True
        )
        if hCenterOuter:
            centers.append(hCenterOuter)

    if centers:
        return centers
    else:
        return None


def findRadicalAxis(circle1, circle2):
    """Calculate the radical axis of two circles.

    On the radical axis (also called power line) of two circles any
    tangents drawn from a point on the axis to both circles have
    the same length.

    http://en.wikipedia.org/wiki/Radical_axis
    http://mathworld.wolfram.com/RadicalLine.html

    See Also
    --------
    findRadicalCenter
    """
    if (geomType(circle1) == "Circle") and (geomType(circle2) == "Circle"):
        print("debug: findRadicalAxis bad parameters! Must be circles.")
        return None

    if DraftVecUtils.equals(circle1.Curve.Center, circle2.Curve.Center):
        return None

    r1 = circle1.Curve.Radius
    r2 = circle1.Curve.Radius
    cen1 = circle1.Curve.Center
    # dist .. the distance from cen1 to cen2.
    dist = DraftVecUtils.dist(cen1, circle2.Curve.Center)
    cenDir = cen1.sub(circle2.Curve.Center)
    cenDir.normalize()

    # Get the perpendicular vector.
    perpCenDir = cenDir.cross(App.Vector(0, 0, 1))
    perpCenDir.normalize()

    # J ... The radical center.
    # K ... The point where the cadical axis crosses the line of cen1->cen2.
    # k1 ... Distance from cen1 to K.
    # k2 ... Distance from cen2 to K.
    # dist = k1 + k2

    k1 = (dist + (r1**2 - r2**2) / dist) / 2.0
    # k2 = dist - k1

    K = App.Vector.add(cen1, cenDir.multiply(k1))

    # K_ .. A point somewhere between K and J; actually with a distance
    # of 1 unit from K.
    K_ = App.Vector.add(K, perpCenDir)

    # Original code didn't specify the value of origin nor dir,
    # so this is a guess:
    # radicalAxis = Part.LineSegment(K, Vector.add(origin, dir))

    origin = App.Vector(0, 0, 0)
    radicalAxis = Part.LineSegment(K, App.Vector.add(origin, perpCenDir))

    if radicalAxis:
        return radicalAxis
    else:
        return None


def findRadicalCenter(circle1, circle2, circle3):
    """Calculate the radical center of three circles.

    It is also called the power center.
    It is the intersection point of the three radical axes of the pairs
    of circles.

    http://en.wikipedia.org/wiki/Power_center_(geometry)
    http://mathworld.wolfram.com/RadicalCenter.html

    See Also
    --------
    findRadicalAxis
    """
    if (
        geomType(circle1) != "Circle"
        or geomType(circle2) != "Circle"
        or geomType(circle3) != "Circle"
    ):
        print("debug: findRadicalCenter bad parameters! Must be circles.")
        return None

    radicalAxis12 = findRadicalAxis(circle1, circle2)
    radicalAxis23 = findRadicalAxis(circle2, circle3)

    if not radicalAxis12 or not radicalAxis23:
        # No radical center could be calculated.
        return None

    intersect = findIntersection(radicalAxis12, radicalAxis23, True, True)

    if intersect:
        return intersect
    else:
        # No radical center could be calculated.
        return None


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2015 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD GbXml exporter"
__author__ = "Yorik van Havre"
__url__ = "https://www.freecad.org"

## @package importGBXML
#  \ingroup ARCH
#  \brief GBXML file format exporter
#
#  This module provides tools to export GBXML files.

from builtins import open as pyopen

import FreeCAD
import Draft

if FreeCAD.GuiUp:
    from draftutils.translate import translate
else:
    # \cond
    def translate(ctx, txt):
        return txt

    # \endcond


def export(objectslist, filename):

    if len(objectslist) != 1:
        FreeCAD.Console.PrintError(
            translate("Arch", "This exporter can currently only export one site object") + "\n"
        )
        return
    site = objectslist[0]
    if Draft.getType(site) != "Site":
        FreeCAD.Console.PrintError(
            translate("Arch", "This exporter can currently only export one site object") + "\n"
        )
        return

    filestream = pyopen(filename, "wb")

    # header
    filestream.write('<?xml version="1.0"?>\n')
    filestream.write(
        "<!-- Exported by FreeCAD %s -->\n" % FreeCAD.Version()[0]
        + FreeCAD.Version()[1]
        + FreeCAD.Version()[2]
    )
    filestream.write("<gbXML\n")
    filestream.write('  xmlns="http://www.gbxml.org/schema"\n')
    filestream.write('  xmlns:xhtml="http://www.w3.org/1999/xhtml"\n')
    filestream.write('  xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"\n')
    filestream.write('  xsi:schemaLocation="http://www.gbxml.org/schema"\n ')
    filestream.write('  temperatureUnit="C"\n')
    filestream.write('  lengthUnit="Meters"\n')
    filestream.write('  areaUnit="SquareMeters"\n')
    filestream.write('  volumeUnit="CubicMeters"\n')
    filestream.write('  useSIUnitsForResults="false" >\n')
    filestream.write("\n")

    # campus
    filestream.write('<Campus id="%s">\n' % site.Name)
    filestream.write("<Location>\n")
    filestream.write("    <ZipcodeOrPostalCode>%s</ZipcodeOrPostalCode>\n" % site.PostalCode)
    filestream.write("    <Longitude>%f</Longitude>\n" % site.Longitude)
    filestream.write("    <Latitude>%f</Latitude>\n" % site.Latitude)
    filestream.write("    <Elevation>%f/Elevation>\n" % site.Elevation.Value)
    filestream.write("    <Name>%s</Name>\n" % site.Label)
    # filestream.write( '    <CADModelAzimuth>0</CADModelAzimuth>\n' )
    # filestream.write( '    <StationId IDType="WMO">53158_2004</StationId>\n' )
    filestream.write("</Location>\n")

    # buildings
    for building in site.Group:
        if Draft.getType(building) == "Building":
            filestream.write(
                '    <Building id="$s" buildingType="$s">\n'
                % (building.Name, building.BuildingType)
            )
            filestream.write("        <Area>$f</Area>\n" % str(building.Area.getValueAs("m^2")))

            # spaces
            for space in Draft.getObjectsOfType(
                Draft.get_group_contents(building.Group, addgroups=True), "Space"
            ):
                if not space.Zone:
                    FreeCAD.Console.PrintError(
                        translate("Arch", "Error: Space '%s' has no Zone. Aborting.") % space.Label
                        + "\n"
                    )
                    return
                filestream.write(
                    '        <Space id="%s" spaceType="%s" zoneIdRef="%s" conditionType="%f">\n'
                    % (space.Name, space.SpaceType, space.Zone.Name, space.Conditioning)
                )
                # filestream.write( '            <CADObjectId>%s</CADObjectId>\n' % space.Name ) # not sure what this is used for?
                filestream.write("            <Name>%s</Name>\n" % space.Label)
                filestream.write("            <Description>%s</Description>\n" % space.Description)
                filestream.write(
                    '            <PeopleNumber unit="NumberOfPeople">%i</PeopleNumber>\n'
                    % space.NumberOfPeople
                )
                filestream.write(
                    '            <LightPowerPerArea unit="WattPerSquareMeter">%f</LightPowerPerArea>\n'
                    % space.LightingPower
                    / space.Area.getValueAs("m^2")
                )
                filestream.write(
                    '            <EquipPowerPerArea unit="WattPerSquareMeter">%f</EquipPowerPerArea>\n'
                    % space.EquipmentPower
                    / space.Area.getValueAs("m^2")
                )
                filestream.write("            <Area>$f</Area>\n" % space.Area.getValueAs("m^2"))
                filestream.write(
                    "            <Volume>$f</Volume>\n"
                    % FreeCAD.Units.Quantity(space.Shape.Volume, FreeCAD.Units.Volume).getValueAs(
                        "m^3"
                    )
                )
                filestream.write('            <ShellGeometry id="%s_geometry">\n' % space.Name)

                # shells
                for solid in space.Shape.Solids:
                    filestream.write("                <ClosedShell>\n")
                    for face in solid.Faces:
                        filestream.write("                    <PolyLoop>\n")
                        for v in face.OuterWire.Vertexes:
                            filestream.write("                        <CartesianPoint>\n")
                            filestream.write(
                                "                            <Coordinate>%f</Coordinate>\n"
                                % v.Point.x
                            )
                            filestream.write(
                                "                            <Coordinate>%f</Coordinate>\n"
                                % v.Point.y
                            )
                            filestream.write(
                                "                            <Coordinate>%f</Coordinate>\n"
                                % v.Point.z
                            )
                            filestream.write("                        </CartesianPoint>\n")
                        filestream.write("                    </PolyLoop>\n")
                    filestream.write("                </ClosedShell>\n")
                filestream.write("            </ShellGeometry>\n")
                filestream.write("        </Space>\n")

                # surfaces
                for i, face in enumerate(space.Shape.Faces):
                    filestream.write(
                        '            <SpaceBoundary isSecondLevelBoundary="false" surfaceIdRef="%s_Face%i"\n'
                        % space.Name,
                        i,
                    )
                    filestream.write("                <PlanarGeometry>\n")
                    filestream.write("                    <PolyLoop>\n")
                    for v in face.OuterWire.Vertexes:
                        filestream.write("                        <CartesianPoint>\n")
                        filestream.write(
                            "                            <Coordinate>%f</Coordinate>\n" % v.Point.x
                        )
                        filestream.write(
                            "                            <Coordinate>%f</Coordinate>\n" % v.Point.y
                        )
                        filestream.write(
                            "                            <Coordinate>%f</Coordinate>\n" % v.Point.z
                        )
                        filestream.write("                        </CartesianPoint>\n")
                    filestream.write("                    </PolyLoop>\n")
                    filestream.write("                </PlanarGeometry>\n")
                    filestream.write("            </SpaceBoundary>\n")

                filestream.write("        </Space>\n")

            filestream.write("    </Building>\n")

    filestream.write("</Campus>\n")

    filestream.write("</gbXML>")


"""
        <Area>18000.00000</Area>
        <Space id="sp1_LabandCorridor_Labcorridor" spaceType="LaboratoryOffice" zoneIdRef="z1_LabandCorridor">
            <Name>Lab corridor</Name>
            <Description/>
            <PeopleNumber unit="NumberOfPeople">1.00000</PeopleNumber>
            <LightPowerPerArea unit="WattPerSquareFoot">1.50000</LightPowerPerArea>
            <EquipPowerPerArea unit="WattPerSquareFoot">0.00000</EquipPowerPerArea>
            <Area>800.00000</Area>
            <Volume>6400.00000</Volume>
            <ShellGeometry id="geo_sp1_LabandCorridor_Labcorridor">
                <ClosedShell>
                    <PolyLoop>
                        <CartesianPoint>
                            <Coordinate>0.00000</Coordinate>
                            <Coordinate>1200.00000</Coordinate>
                            <Coordinate>0.00000</Coordinate>
                        </CartesianPoint>
                        <CartesianPoint>
                            <Coordinate>0.00000</Coordinate>
                            <Coordinate>1200.00000</Coordinate>
                            <Coordinate>96.00000</Coordinate>
                        </CartesianPoint>
                        <CartesianPoint>
                            <Coordinate>480.00000</Coordinate>
                            <Coordinate>1200.00000</Coordinate>
                            <Coordinate>96.00000</Coordinate>
                        </CartesianPoint>
                        <CartesianPoint>
                            <Coordinate>480.00000</Coordinate>
                            <Coordinate>1200.00000</Coordinate>
                            <Coordinate>0.00000</Coordinate>
                        </CartesianPoint>
                    </PolyLoop>

                    ... repeat

                </ClosedShell>
            </ShellGeometry>

            <SpaceBoundary isSecondLevelBoundary="false" surfaceIdRef="aim1095">
              <PlanarGeometry>
                <PolyLoop>
                  <CartesianPoint>
                    <Coordinate>9.981497</Coordinate>
                    <Coordinate>-31.19363</Coordinate>
                    <Coordinate>0</Coordinate>
                  </CartesianPoint>
                  <CartesianPoint>
                    <Coordinate>9.981497</Coordinate>
                    <Coordinate>-5.193626</Coordinate>
                    <Coordinate>0</Coordinate>
                  </CartesianPoint>
                  <CartesianPoint>
                    <Coordinate>9.981497</Coordinate>
                    <Coordinate>-5.193626</Coordinate>
                    <Coordinate>100</Coordinate>
                  </CartesianPoint>
                  <CartesianPoint>
                    <Coordinate>9.981497</Coordinate>
                    <Coordinate>-31.19363</Coordinate>
                    <Coordinate>100</Coordinate>
                  </CartesianPoint>
                </PolyLoop>
              </PlanarGeometry>
            </SpaceBoundary>




            <CADObjectId>21E2</CADObjectId>
        </Space>

        ... repeat

    </Building>



    <Surface id="su1_Floor" surfaceType="UndergroundSlab" constructionIdRef="construction-1">
        <Name>Floor</Name>
        <AdjacentSpaceId spaceIdRef="sp1_LabandCorridor_Labcorridor"/>
        <RectangularGeometry>
            <Azimuth>90.00</Azimuth>
            <CartesianPoint>
                <Coordinate>0.00000</Coordinate>
                <Coordinate>1320.00000</Coordinate>
                <Coordinate>0.00000</Coordinate>
            </CartesianPoint>
            <Tilt>180.00</Tilt>
            <Height>480.00000</Height>
            <Width>240.00000</Width>
        </RectangularGeometry>
        <PlanarGeometry>
            <PolyLoop>
                <CartesianPoint>
                    <Coordinate>0.00000</Coordinate>
                    <Coordinate>1320.00000</Coordinate>
                    <Coordinate>0.00000</Coordinate>
                </CartesianPoint>
                <CartesianPoint>
                    <Coordinate>480.00000</Coordinate>
                    <Coordinate>1320.00000</Coordinate>
                    <Coordinate>0.00000</Coordinate>
                </CartesianPoint>
                <CartesianPoint>
                    <Coordinate>960.00000</Coordinate>
                    <Coordinate>1320.00000</Coordinate>
                    <Coordinate>0.00000</Coordinate>
                </CartesianPoint>
                <CartesianPoint>
                    <Coordinate>960.00000</Coordinate>
                    <Coordinate>1200.00000</Coordinate>
                    <Coordinate>0.00000</Coordinate>
                </CartesianPoint>
                <CartesianPoint>
                    <Coordinate>480.00000</Coordinate>
                    <Coordinate>1200.00000</Coordinate>
                    <Coordinate>0.00000</Coordinate>
                </CartesianPoint>
                <CartesianPoint>
                    <Coordinate>0.00000</Coordinate>
                    <Coordinate>1200.00000</Coordinate>
                    <Coordinate>0.00000</Coordinate>
                </CartesianPoint>
            </PolyLoop>
        </PlanarGeometry>
    </Surface>



    <Surface id="su44_Surface4" surfaceType="ExteriorWall" constructionIdRef="construction-3">
        <Name>Surface 4</Name>
        <AdjacentSpaceId spaceIdRef="sp7_Office_Office6"/>
        <RectangularGeometry>
            <Azimuth>180.00</Azimuth>
            <CartesianPoint>
                <Coordinate>960.00000</Coordinate>
                <Coordinate>0.00000</Coordinate>
                <Coordinate>0.00000</Coordinate>
            </CartesianPoint>
            <Tilt>90.00</Tilt>
            <Height>114.00000</Height>
            <Width>480.00000</Width>
        </RectangularGeometry>
        <PlanarGeometry>
            <PolyLoop>
                <CartesianPoint>
                    <Coordinate>960.00000</Coordinate>
                    <Coordinate>0.00000</Coordinate>
                    <Coordinate>0.00000</Coordinate>
                </CartesianPoint>
                <CartesianPoint>
                    <Coordinate>1440.00000</Coordinate>
                    <Coordinate>0.00000</Coordinate>
                    <Coordinate>0.00000</Coordinate>
                </CartesianPoint>
                <CartesianPoint>
                    <Coordinate>1440.00000</Coordinate>
                    <Coordinate>0.00000</Coordinate>
                    <Coordinate>114.00000</Coordinate>
                </CartesianPoint>
                <CartesianPoint>
                    <Coordinate>960.00000</Coordinate>
                    <Coordinate>0.00000</Coordinate>
                    <Coordinate>114.00000</Coordinate>
                </CartesianPoint>
            </PolyLoop>
        </PlanarGeometry>
        <Opening id="su44-op1_Opening1" openingType="OperableWindow" windowTypeIdRef="windowType-1">
            <Name>Opening1</Name>
            <RectangularGeometry>
                <CartesianPoint>
                    <Coordinate>96.00000</Coordinate>
                    <Coordinate>24.00000</Coordinate>
                </CartesianPoint>
                <Height>72.00000</Height>
                <Width>48.00000</Width>
            </RectangularGeometry>
            <PlanarGeometry>
                <PolyLoop>
                    <CartesianPoint>
                        <Coordinate>1056.00000</Coordinate>
                        <Coordinate>0.00000</Coordinate>
                        <Coordinate>24.00000</Coordinate>
                    </CartesianPoint>
                    <CartesianPoint>
                        <Coordinate>1104.00000</Coordinate>
                        <Coordinate>0.00000</Coordinate>
                        <Coordinate>24.00000</Coordinate>
                    </CartesianPoint>
                    <CartesianPoint>
                        <Coordinate>1104.00000</Coordinate>
                        <Coordinate>0.00000</Coordinate>
                        <Coordinate>96.00000</Coordinate>
                    </CartesianPoint>
                    <CartesianPoint>
                        <Coordinate>1056.00000</Coordinate>
                        <Coordinate>0.00000</Coordinate>
                        <Coordinate>96.00000</Coordinate>
                    </CartesianPoint>
                </PolyLoop>
            </PlanarGeometry>
        </Opening>
        <Opening id="su44-op2_Opening2" openingType="OperableWindow" windowTypeIdRef="windowType-1">
            <Name>Opening2</Name>
            <RectangularGeometry>
                <CartesianPoint>
                    <Coordinate>216.00000</Coordinate>
                    <Coordinate>24.00000</Coordinate>
                </CartesianPoint>
                <Height>72.00000</Height>
                <Width>48.00000</Width>
            </RectangularGeometry>
            <PlanarGeometry>
                <PolyLoop>
                    <CartesianPoint>
                        <Coordinate>1176.00000</Coordinate>
                        <Coordinate>0.00000</Coordinate>
                        <Coordinate>24.00000</Coordinate>
                    </CartesianPoint>
                    <CartesianPoint>
                        <Coordinate>1224.00000</Coordinate>
                        <Coordinate>0.00000</Coordinate>
                        <Coordinate>24.00000</Coordinate>
                    </CartesianPoint>
                    <CartesianPoint>
                        <Coordinate>1224.00000</Coordinate>
                        <Coordinate>0.00000</Coordinate>
                        <Coordinate>96.00000</Coordinate>
                    </CartesianPoint>
                    <CartesianPoint>
                        <Coordinate>1176.00000</Coordinate>
                        <Coordinate>0.00000</Coordinate>
                        <Coordinate>96.00000</Coordinate>
                    </CartesianPoint>
                </PolyLoop>
            </PlanarGeometry>
        </Opening>
        <Opening id="su44-op3_Opening3" openingType="OperableWindow" windowTypeIdRef="windowType-1">
            <Name>Opening3</Name>
            <RectangularGeometry>
                <CartesianPoint>
                    <Coordinate>336.00000</Coordinate>
                    <Coordinate>24.00000</Coordinate>
                </CartesianPoint>
                <Height>72.00000</Height>
                <Width>48.00000</Width>
            </RectangularGeometry>
            <PlanarGeometry>
                <PolyLoop>
                    <CartesianPoint>
                        <Coordinate>1296.00000</Coordinate>
                        <Coordinate>0.00000</Coordinate>
                        <Coordinate>24.00000</Coordinate>
                    </CartesianPoint>
                    <CartesianPoint>
                        <Coordinate>1344.00000</Coordinate>
                        <Coordinate>0.00000</Coordinate>
                        <Coordinate>24.00000</Coordinate>
                    </CartesianPoint>
                    <CartesianPoint>
                        <Coordinate>1344.00000</Coordinate>
                        <Coordinate>0.00000</Coordinate>
                        <Coordinate>96.00000</Coordinate>
                    </CartesianPoint>
                    <CartesianPoint>
                        <Coordinate>1296.00000</Coordinate>
                        <Coordinate>0.00000</Coordinate>
                        <Coordinate>96.00000</Coordinate>
                    </CartesianPoint>
                </PolyLoop>
            </PlanarGeometry>
        </Opening>
    </Surface>


    ... repeat

</Campus>



<Construction id="construction-1">
    <Name>Standard</Name>
    <Description/>
</Construction>
<Construction id="construction-2">
    <Name>Standard</Name>
    <Description/>
</Construction>
<Construction id="construction-3">
    <Name>Standard</Name>
    <Description/>
</Construction>
<WindowType id="windowType-1">
    <Name>Standard</Name>
    <Description/>
</WindowType>



<Zone id="z1_LabandCorridor">
    <Name>Lab and Corridor</Name>
    <Description/>
    <AirChangesPerHour>0"</AirChangesPerHour>
    <FlowPerArea unit="CFMPerSquareFoot">0.00000</FlowPerArea>
    <FlowPerPerson unit="CFM">0.00000</FlowPerPerson>
    <OAFlowPerArea unit="CFMPerSquareFoot">2.37037</OAFlowPerArea>
    <OAFlowPerPerson unit="CFM">812.69841</OAFlowPerPerson>
    <DesignHeatT>72.00000</DesignHeatT>
    <DesignCoolT>75.00000</DesignCoolT>
</Zone>
<Zone id="z2_Office">
    <Name>Office</Name>
    <Description/>
    <AirChangesPerHour>1"</AirChangesPerHour>
    <FlowPerArea unit="CFMPerSquareFoot">0.13333</FlowPerArea>
    <FlowPerPerson unit="CFM">20.00000</FlowPerPerson>
    <OAFlowPerArea unit="CFMPerSquareFoot">0.05333</OAFlowPerArea>
    <OAFlowPerPerson unit="CFM">8.00000</OAFlowPerPerson>
    <DesignHeatT>72.00000</DesignHeatT>
    <DesignCoolT>75.00000</DesignCoolT>
</Zone>
<Zone id="z3_Warehouse">
    <Name>Warehouse</Name>
    <Description/>
    <AirChangesPerHour>5/32"</AirChangesPerHour>
    <FlowPerArea unit="CFMPerSquareFoot">0.05000</FlowPerArea>
    <FlowPerPerson unit="CFM">25.71429</FlowPerPerson>
    <OAFlowPerArea unit="CFMPerSquareFoot">0.00000</OAFlowPerArea>
    <OAFlowPerPerson unit="CFM">0.00000</OAFlowPerPerson>
    <DesignHeatT>60.00000</DesignHeatT>
    <DesignCoolT>80.00000</DesignCoolT>
</Zone>

<DocumentHistory>
    <ProgramInfo id="adesk-rvt-1">
        <CompanyName>Autodesk, Inc.</CompanyName>
        <ProductName>Autodesk Project Vasari CEA</ProductName>
        <Version>TP2.0 20110514_1800</Version>
        <Platform>Microsoft Windows XP</Platform>
    </ProgramInfo>
</DocumentHistory>

<Results xmlns="" id="sp3_LabandCorridor_Lab1" objectIdRef="sp3_LabandCorridor_Lab1" resultsType="CoolingLoad" unit="BtuPerHour">
    <ObjectId>sp3_LabandCorridor_Lab1</ObjectId>
    <Value>5534.837890625</Value>
    <Description>Space Cooling Roof Cond</Description>
    <CADObjectId>21E3</CADObjectId>
</Results>

... repeat

</gbXML>"""



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2016 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************


import FreeCAD
import Path
import Path.Op.PocketBase as PathPocketBase
import PathScripts.PathUtils as PathUtils
from PySide.QtCore import QT_TRANSLATE_NOOP
import numpy

# lazily loaded modules
from lazy_loader.lazy_loader import LazyLoader

Part = LazyLoader("Part", globals(), "Part")

__title__ = "CAM Mill Face Operation"
__author__ = "sliptonic (Brad Collette)"
__url__ = "https://www.freecad.org"
__doc__ = "Class and implementation of Mill Facing operation."
__contributors__ = "russ4262 (Russell Johnson)"


if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())

translate = FreeCAD.Qt.translate


class ObjectFace(PathPocketBase.ObjectPocket):
    """Proxy object for Mill Facing operation."""

    @classmethod
    def propertyEnumerations(self, dataType="data"):
        """helixOpPropertyEnumerations(dataType="data")... return property enumeration lists of specified dataType.
        Args:
            dataType = 'data', 'raw', 'translated'
        Notes:
        'data' is list of internal string literals used in code
        'raw' is list of (translated_text, data_string) tuples
        'translated' is list of translated string literals
        """

        enums = {
            "BoundaryShape": [
                (translate("CAM_Pocket", "Boundbox"), "Boundbox"),
                (translate("CAM_Pocket", "Face Region"), "Face Region"),
                (translate("CAM_Pocket", "Perimeter"), "Perimeter"),
                (translate("CAM_Pocket", "Stock"), "Stock"),
            ],
        }

        if dataType == "raw":
            return enums

        data = list()
        idx = 0 if dataType == "translated" else 1

        Path.Log.debug(enums)

        for k, v in enumerate(enums):
            data.append((v, [tup[idx] for tup in enums[v]]))
        Path.Log.debug(data)

        return data

    def initPocketOp(self, obj):
        Path.Log.track()
        """initPocketOp(obj) ... create facing specific properties"""
        obj.addProperty(
            "App::PropertyEnumeration",
            "BoundaryShape",
            "Face",
            QT_TRANSLATE_NOOP("App::Property", "Shape to use for calculating Boundary"),
        )
        obj.addProperty(
            "App::PropertyBool",
            "ClearEdges",
            "Face",
            QT_TRANSLATE_NOOP(
                "App::Property", "Clear edges of surface (Only applicable to BoundBox)"
            ),
        )
        if not hasattr(obj, "ExcludeRaisedAreas"):
            obj.addProperty(
                "App::PropertyBool",
                "ExcludeRaisedAreas",
                "Face",
                QT_TRANSLATE_NOOP("App::Property", "Exclude milling raised areas inside the face."),
            )

        for n in self.propertyEnumerations():
            setattr(obj, n[0], n[1])

    def pocketInvertExtraOffset(self):
        return True

    def areaOpOnChanged(self, obj, prop):
        """areaOpOnChanged(obj, prop) ... facing specific depths calculation."""
        Path.Log.track(prop)
        if prop == "StepOver" and obj.StepOver == 0:
            obj.StepOver = 1

        # default depths calculation not correct for facing
        if prop == "Base":
            job = PathUtils.findParentJob(obj)
            if job:
                obj.OpStartDepth = job.Stock.Shape.BoundBox.ZMax

            if len(obj.Base) >= 1:
                Path.Log.debug("processing")
                sublist = []
                for i in obj.Base:
                    o = i[0]
                    for s in i[1]:
                        sublist.append(o.Shape.getElement(s))

                # If the operation has a geometry identified the Finaldepth
                # is the top of the bboundbox which includes all features.
                # Otherwise, top of part.

                obj.OpFinalDepth = Part.makeCompound(sublist).BoundBox.ZMax
            elif job:
                obj.OpFinalDepth = job.Proxy.modelBoundBox(job).ZMax

    def areaOpShapes(self, obj):
        """areaOpShapes(obj) ... return top face"""
        # Facing is done either against base objects
        self.removalshapes = []
        holeShape = None

        Path.Log.debug("depthparams: {}".format([i for i in self.depthparams]))

        if obj.Base:
            Path.Log.debug("obj.Base: {}".format(obj.Base))
            faces = []
            holes = []
            holeEnvs = []
            oneBase = [obj.Base[0][0], True]
            sub0 = getattr(obj.Base[0][0].Shape, obj.Base[0][1][0])
            minHeight = sub0.BoundBox.ZMax

            for b in obj.Base:
                for sub in b[1]:
                    shape = getattr(b[0].Shape, sub)
                    if isinstance(shape, Part.Face):
                        faces.append(shape)
                        if shape.BoundBox.ZMin < minHeight:
                            minHeight = shape.BoundBox.ZMin
                        # Limit to one model base per operation
                        if oneBase[0] is not b[0]:
                            oneBase[1] = False
                        if numpy.isclose(abs(shape.normalAt(0, 0).z), 1):  # horizontal face
                            # Analyze internal closed wires to determine if raised or a recess
                            for wire in shape.Wires[1:]:
                                if obj.ExcludeRaisedAreas:
                                    ip = self.isPocket(b[0], shape, wire)
                                    if ip is False:
                                        holes.append((b[0].Shape, wire))
                                else:
                                    holes.append((b[0].Shape, wire))
                    else:
                        Path.Log.warning(
                            'The base subobject, "{0}," is not a face. Ignoring "{0}."'.format(sub)
                        )

            if obj.ExcludeRaisedAreas and len(holes) > 0:
                for shape, wire in holes:
                    f = Part.makeFace(wire, "Part::FaceMakerSimple")
                    env = PathUtils.getEnvelope(shape, subshape=f, depthparams=self.depthparams)
                    holeEnvs.append(env)
                    holeShape = Part.makeCompound(holeEnvs)

            Path.Log.debug("Working on a collection of faces {}".format(faces))
            planeshape = Part.makeCompound(faces)

        # If no base object, do planing of top surface of entire model
        else:
            planeshape = Part.makeCompound([base.Shape for base in self.model])
            Path.Log.debug("Working on a shape {}".format(obj.Label))

        # Find the correct shape depending on Boundary shape.
        Path.Log.debug("Boundary Shape: {}".format(obj.BoundaryShape))
        bb = planeshape.BoundBox

        # Apply offset for clearing edges
        offset = 0
        if obj.ClearEdges:
            offset = self.radius + 0.1

        bb.XMin = bb.XMin - offset
        bb.YMin = bb.YMin - offset
        bb.XMax = bb.XMax + offset
        bb.YMax = bb.YMax + offset

        if obj.BoundaryShape == "Boundbox":
            bbperim = Part.makeBox(
                bb.XLength,
                bb.YLength,
                1,
                FreeCAD.Vector(bb.XMin, bb.YMin, bb.ZMin),
                FreeCAD.Vector(0, 0, 1),
            )
            env = PathUtils.getEnvelope(partshape=bbperim, depthparams=self.depthparams)
            if obj.ExcludeRaisedAreas and oneBase[1]:
                includedFaces = self.getAllIncludedFaces(oneBase[0], env, faceZ=minHeight)
                if len(includedFaces) > 0:
                    includedShape = Part.makeCompound(includedFaces)
                    includedEnv = PathUtils.getEnvelope(
                        oneBase[0].Shape,
                        subshape=includedShape,
                        depthparams=self.depthparams,
                    )
                    env = env.cut(includedEnv)
        elif obj.BoundaryShape == "Stock":
            stock = PathUtils.findParentJob(obj).Stock.Shape
            env = stock

            if obj.ExcludeRaisedAreas and oneBase[1]:
                includedFaces = self.getAllIncludedFaces(oneBase[0], stock, faceZ=minHeight)
                if len(includedFaces) > 0:
                    stockEnv = PathUtils.getEnvelope(partshape=stock, depthparams=self.depthparams)
                    includedShape = Part.makeCompound(includedFaces)
                    includedEnv = PathUtils.getEnvelope(
                        oneBase[0].Shape,
                        subshape=includedShape,
                        depthparams=self.depthparams,
                    )
                    env = stockEnv.cut(includedEnv)
        elif obj.BoundaryShape == "Perimeter":
            if obj.ClearEdges:
                psZMin = planeshape.BoundBox.ZMin
                ofstShape = PathUtils.getOffsetArea(
                    planeshape, self.radius * 1.25, plane=planeshape
                )
                ofstShape.translate(FreeCAD.Vector(0.0, 0.0, psZMin - ofstShape.BoundBox.ZMin))
                env = PathUtils.getEnvelope(partshape=ofstShape, depthparams=self.depthparams)
            else:
                env = PathUtils.getEnvelope(partshape=planeshape, depthparams=self.depthparams)
        elif obj.BoundaryShape == "Face Region":
            baseShape = planeshape  # oneBase[0].Shape
            psZMin = planeshape.BoundBox.ZMin
            ofst = 0.0
            if obj.ClearEdges:
                ofst = self.tool.Diameter * 0.51
            ofstShape = PathUtils.getOffsetArea(planeshape, ofst, plane=planeshape)
            ofstShape.translate(FreeCAD.Vector(0.0, 0.0, psZMin - ofstShape.BoundBox.ZMin))

            # Calculate custom depth params for removal shape envelope, with start and final depth buffers
            custDepthparams = self._customDepthParams(
                obj, obj.StartDepth.Value + 0.2, obj.FinalDepth.Value - 0.1
            )  # only an envelope
            ofstShapeEnv = PathUtils.getEnvelope(partshape=ofstShape, depthparams=custDepthparams)
            if obj.ExcludeRaisedAreas:
                env = ofstShapeEnv.cut(baseShape)
                env.translate(
                    FreeCAD.Vector(0.0, 0.0, -0.00001)
                )  # lower removal shape into buffer zone
            else:
                env = ofstShapeEnv

        if holeShape:
            Path.Log.debug("Processing holes and face ...")
            holeEnv = PathUtils.getEnvelope(partshape=holeShape, depthparams=self.depthparams)
            newEnv = env.cut(holeEnv)
            tup = newEnv, False, "pathMillFace"
        else:
            Path.Log.debug("Processing solid face ...")
            tup = env, False, "pathMillFace"

        self.removalshapes.append(tup)
        obj.removalshape = self.removalshapes[0][0]  # save removal shape

        return self.removalshapes

    def areaOpSetDefaultValues(self, obj, job):
        """areaOpSetDefaultValues(obj, job) ... initialize mill facing properties"""
        obj.StepOver = 50
        obj.ZigZagAngle = 45.0
        obj.ExcludeRaisedAreas = False
        obj.ClearEdges = False

        # need to overwrite the default depth calculations for facing
        if job and len(job.Model.Group) > 0:
            obj.OpStartDepth = job.Stock.Shape.BoundBox.ZMax
            obj.OpFinalDepth = job.Proxy.modelBoundBox(job).ZMax

            # If the operation has a geometry identified the Finaldepth
            # is the top of the boundbox which includes all features.
            if len(obj.Base) >= 1:
                shapes = []
                for base, subs in obj.Base:
                    for s in subs:
                        shapes.append(getattr(base.Shape, s))
                obj.OpFinalDepth = Part.makeCompound(shapes).BoundBox.ZMax

    def isPocket(self, b, f, w):
        e = w.Edges[0]
        for fi in range(0, len(b.Shape.Faces)):
            face = b.Shape.Faces[fi]
            for ei in range(0, len(face.Edges)):
                edge = face.Edges[ei]
                if e.isSame(edge):
                    if f is face:
                        # Alternative: run loop to see if all edges are same
                        pass  # same source face, look for another
                    else:
                        if face.CenterOfMass.z < f.CenterOfMass.z:
                            return True
        return False

    def getAllIncludedFaces(self, base, env, faceZ):
        """getAllIncludedFaces(base, env, faceZ)...
        Return all `base` faces extending above `faceZ` whose boundboxes overlap with the `env` boundbox.
        """
        included = []

        eXMin = env.BoundBox.XMin
        eXMax = env.BoundBox.XMax
        eYMin = env.BoundBox.YMin
        eYMax = env.BoundBox.YMax
        eZMin = faceZ

        def isOverlap(faceMin, faceMax, envMin, envMax):
            if faceMax > envMin:
                if faceMin <= envMax or faceMin == envMin:
                    return True
            return False

        for fi in range(0, len(base.Shape.Faces)):
            # Check all faces of `base` shape
            incl = False
            face = base.Shape.Faces[fi]
            fXMin = face.BoundBox.XMin
            fXMax = face.BoundBox.XMax
            fYMin = face.BoundBox.YMin
            fYMax = face.BoundBox.YMax
            fZMax = face.BoundBox.ZMax

            if fZMax > eZMin:
                # Include face if its boundbox overlaps envelope boundbox
                if isOverlap(fXMin, fXMax, eXMin, eXMax):  # check X values
                    if isOverlap(fYMin, fYMax, eYMin, eYMax):  # check Y values
                        incl = True
            if incl:
                included.append(face)
        return included


def SetupProperties():
    setup = PathPocketBase.SetupProperties()
    setup.append("BoundaryShape")
    setup.append("ExcludeRaisedAreas")
    setup.append("ClearEdges")
    return setup


def Create(name, obj=None, parentJob=None):
    """Create(name) ... Creates and returns a Mill Facing operation."""
    if obj is None:
        obj = FreeCAD.ActiveDocument.addObject("Path::FeaturePython", name)
    obj.Proxy = ObjectFace(obj, name, parentJob)
    return obj



# SPDX-License-Identifier: LGPL-2.1-or-later
# /**************************************************************************
#                                                                           *
#    Copyright (c) 2023 Ondsel <development@ondsel.com>                     *
#                                                                           *
#    This file is part of FreeCAD.                                          *
#                                                                           *
#    FreeCAD is free software: you can redistribute it and/or modify it     *
#    under the terms of the GNU Lesser General Public License as            *
#    published by the Free Software Foundation, either version 2.1 of the   *
#    License, or (at your option) any later version.                        *
#                                                                           *
#    FreeCAD is distributed in the hope that it will be useful, but         *
#    WITHOUT ANY WARRANTY; without even the implied warranty of             *
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
#    Lesser General Public License for more details.                        *
#                                                                           *
#    You should have received a copy of the GNU Lesser General Public       *
#    License along with FreeCAD. If not, see                                *
#    <https://www.gnu.org/licenses/>.                                       *
#                                                                           *
# **************************************************************************/

import FreeCAD as App
import UtilsAssembly

from PySide.QtCore import QT_TRANSLATE_NOOP
from PySide.QtWidgets import QFileDialog

if App.GuiUp:
    import FreeCADGui as Gui


__title__ = "Assembly Command Create Assembly"
__author__ = "Ondsel"
__url__ = "https://www.freecad.org"


class CommandExportASMT:
    def __init__(self):
        pass

    def GetResources(self):
        return {
            "Pixmap": "Assembly_ExportASMT",
            "MenuText": QT_TRANSLATE_NOOP("Assembly_ExportASMT", "Export ASMT File"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Assembly_ExportASMT",
                "Export currently active assembly as a ASMT file.",
            ),
            "CmdType": "ForEdit",
        }

    def IsActive(self):
        return UtilsAssembly.isAssemblyCommandActive()

    def Activated(self):
        document = App.ActiveDocument
        if not document:
            return

        assembly = UtilsAssembly.activeAssembly()
        if not assembly:
            return

        # Prompt the user for a file location and name
        defaultFileName = document.Name + ".asmt"
        filePath, _ = QFileDialog.getSaveFileName(
            None,
            "Save ASMT File",
            defaultFileName,
            "ASMT Files (*.asmt);;All Files (*)",
        )

        if filePath:
            Gui.addModule("UtilsAssembly")
            Gui.doCommand("assembly = UtilsAssembly.activeAssembly()")
            Gui.doCommand(f'assembly.exportAsASMT("{filePath}")')


if App.GuiUp:
    Gui.addCommand("Assembly_ExportASMT", CommandExportASMT())



# SPDX-License-Identifier: LGPL-2.1-or-later

import unittest
import pathlib
import asyncio
import tempfile
from uuid import uuid4
from Path.Tool.assets import (
    AssetUri,
    AssetStore,
    MemoryStore,
    FileStore,
)


class BaseTestPathToolAssetStore(unittest.TestCase):
    """
    Base test suite for Path Tool Asset Stores assuming full versioning support.
    Store-agnostic tests without direct file system access.
    """

    store: AssetStore

    def setUp(self):
        self.tmp_dir = tempfile.TemporaryDirectory()
        self.tmp_path = pathlib.Path(self.tmp_dir.name)

    def tearDown(self):
        self.tmp_dir.cleanup()

    def test_name(self):
        self.assertIsNotNone(self.store)
        self.assertIsInstance(self.store.name, str)
        self.assertTrue(len(self.store.name) > 0)

    def test_create_and_get(self):
        async def async_test():
            data = b"test data"
            asset_type = f"type_{uuid4()}"
            asset_id = f"asset_{uuid4()}"
            uri = await self.store.create(asset_type, asset_id, data)

            self.assertIsInstance(uri, AssetUri)
            self.assertEqual(uri.asset_type, asset_type)
            self.assertEqual(uri.asset_id, asset_id)
            self.assertIsNotNone(uri.version)

            retrieved_data = await self.store.get(uri)
            self.assertEqual(retrieved_data, data)

            # Test non-existent URI
            non_existent_uri = AssetUri.build(
                asset_type="non_existent", asset_id="missing", version="1"
            )
            with self.assertRaises(FileNotFoundError):
                await self.store.get(non_existent_uri)

        asyncio.run(async_test())

    def test_delete(self):
        async def async_test():
            data = b"data to delete"
            asset_type = "delete_type"
            asset_id = f"asset_{uuid4()}"
            uri = await self.store.create(asset_type, asset_id, data)

            await self.store.delete(uri)

            with self.assertRaises(FileNotFoundError):
                await self.store.get(uri)

            # Deleting non-existent URI should not raise
            non_existent_uri = AssetUri.build(
                asset_type="non_existent", asset_id="missing", version="1"
            )
            await self.store.delete(non_existent_uri)

        asyncio.run(async_test())

    def test_is_empty(self):
        async def async_test():
            self.assertTrue(await self.store.is_empty())
            self.assertTrue(await self.store.is_empty("type1"))

            uri1 = await self.store.create("type1", f"asset_{uuid4()}", b"data")
            self.assertFalse(await self.store.is_empty())
            self.assertFalse(await self.store.is_empty("type1"))
            self.assertTrue(await self.store.is_empty("type2"))

            uri2 = await self.store.create("type2", f"asset_{uuid4()}", b"data")
            await self.store.delete(uri1)
            self.assertFalse(await self.store.is_empty())
            self.assertTrue(await self.store.is_empty("type1"))
            self.assertFalse(await self.store.is_empty("type2"))

            await self.store.delete(uri2)
            self.assertTrue(await self.store.is_empty())

        asyncio.run(async_test())

    def test_count_assets(self):
        async def async_test():
            self.assertEqual(await self.store.count_assets(), 0)
            self.assertEqual(await self.store.count_assets("type1"), 0)

            uri1 = await self.store.create("type1", f"asset1_{uuid4()}", b"data1")
            self.assertEqual(await self.store.count_assets(), 1)
            self.assertEqual(await self.store.count_assets("type1"), 1)
            self.assertEqual(await self.store.count_assets("type2"), 0)

            uri2 = await self.store.create("type2", f"asset2_{uuid4()}", b"data2")
            uri3 = await self.store.create("type1", f"asset3_{uuid4()}", b"data3")
            self.assertEqual(await self.store.count_assets(), 3)
            self.assertEqual(await self.store.count_assets("type1"), 2)
            self.assertEqual(await self.store.count_assets("type2"), 1)

            await self.store.delete(uri1)
            self.assertEqual(await self.store.count_assets(), 2)
            self.assertEqual(await self.store.count_assets("type1"), 1)
            self.assertEqual(await self.store.count_assets("type2"), 1)

            await self.store.delete(uri2)
            await self.store.delete(uri3)
            self.assertEqual(await self.store.count_assets(), 0)
            self.assertEqual(await self.store.count_assets("type1"), 0)
            self.assertEqual(await self.store.count_assets("type2"), 0)

        asyncio.run(async_test())

    def test_list_assets(self):
        async def async_test():
            asset_typedata = [
                ("type1", f"asset1_{uuid4()}", b"data1"),
                ("type1", f"asset2_{uuid4()}", b"data2"),
                ("type2", f"asset3_{uuid4()}", b"data3"),
            ]

            uris = []
            for asset_type, asset_id, data in asset_typedata:
                uri = await self.store.create(asset_type, asset_id, data)
                uris.append(uri)

            all_assets = await self.store.list_assets()
            self.assertEqual(len(all_assets), 3)
            for uri in uris:
                self.assertTrue(any(u.asset_id == uri.asset_id for u in all_assets))

            type1_assets = await self.store.list_assets(asset_type="type1")
            self.assertEqual(len(type1_assets), 2)
            self.assertTrue(any(u.asset_id == uris[0].asset_id for u in type1_assets))
            self.assertTrue(any(u.asset_id == uris[1].asset_id for u in type1_assets))

            paginated = await self.store.list_assets(limit=2)
            self.assertEqual(len(paginated), 2)

        asyncio.run(async_test())

    def test_update_versioning(self):
        async def async_test():
            initial_data = b"initial data"
            updated_data = b"updated data"
            asset_type = f"update_{uuid4()}"
            asset_id = f"asset_{uuid4()}"

            uri1 = await self.store.create(asset_type, asset_id, initial_data)
            uri2 = await self.store.update(uri1, updated_data)

            self.assertEqual(uri1.asset_type, uri2.asset_type)
            self.assertEqual(uri1.asset_id, uri2.asset_id)
            self.assertEqual(uri1.version, "1")
            self.assertEqual(uri2.version, "2")
            self.assertNotEqual(uri1.version, uri2.version)

            self.assertEqual(await self.store.get(uri1), initial_data)
            self.assertEqual(await self.store.get(uri2), updated_data)

            with self.assertRaises(FileNotFoundError):
                non_existent_uri = AssetUri.build(
                    asset_type="non_existent", asset_id="missing", version="1"
                )
                await self.store.update(non_existent_uri, b"data")

        asyncio.run(async_test())

    def test_list_versions(self):
        async def async_test():
            asset_type = f"version_{uuid4()}"
            asset_id = f"asset_{uuid4()}"
            data1 = b"version1"
            data2 = b"version2"

            uri1 = await self.store.create(asset_type, asset_id, data1)
            uri2 = await self.store.update(uri1, data2)

            versions = await self.store.list_versions(uri1)
            self.assertEqual(len(versions), 2)
            version_ids = {v.version for v in versions if v.version}
            self.assertEqual(version_ids, {uri1.version, uri2.version})

            non_existent_uri = AssetUri.build(
                asset_type="non_existent", asset_id="missing", version="1"
            )
            self.assertEqual(await self.store.list_versions(non_existent_uri), [])

        asyncio.run(async_test())

    def test_create_with_empty_data(self):
        async def async_test():
            data = b""
            asset_type = "shape"
            asset_id = f"asset_{uuid4()}"
            uri = await self.store.create(asset_type, asset_id, data)

            self.assertIsInstance(uri, AssetUri)
            retrieved_data = await self.store.get(uri)
            self.assertEqual(retrieved_data, data)

        asyncio.run(async_test())

    def test_list_assets_non_existent_type(self):
        async def async_test():
            assets = await self.store.list_assets(asset_type=f"non_existent_type_{uuid4()}")
            self.assertEqual(len(assets), 0)

        asyncio.run(async_test())

    def test_list_assets_pagination_offset_too_high(self):
        async def async_test():
            await self.store.create("shape", f"asset1_{uuid4()}", b"data")
            assets = await self.store.list_assets(offset=100)  # Assuming less than 100 assets
            self.assertEqual(len(assets), 0)

        asyncio.run(async_test())

    def test_list_assets_pagination_limit_zero(self):
        async def async_test():
            await self.store.create("shape", f"asset1_{uuid4()}", b"data")
            assets = await self.store.list_assets(limit=0)
            self.assertEqual(len(assets), 0)

        asyncio.run(async_test())

    def test_create_delete_recreate(self):
        async def async_test():
            asset_type = "shape"
            asset_id = f"asset_{uuid4()}"
            data1 = b"first data"
            data2 = b"second data"

            uri1 = await self.store.create(asset_type, asset_id, data1)
            self.assertEqual(await self.store.get(uri1), data1)
            # For versioned stores, this would be version "1"
            # For stores that don't deeply track versions, it's important what happens on recreate

            await self.store.delete(uri1)
            with self.assertRaises(FileNotFoundError):
                await self.store.get(uri1)

            uri2 = await self.store.create(asset_type, asset_id, data2)
            self.assertEqual(await self.store.get(uri2), data2)

            # Behavior of uri1.version vs uri2.version depends on store implementation
            # For a fully versioned store that starts fresh:
            self.assertEqual(
                uri2.version, "1", "Recreating should yield version 1 for a fresh start"
            )

            # Ensure only the new asset exists if the store fully removes old versions
            versions = await self.store.list_versions(
                AssetUri.build(asset_type=asset_type, asset_id=asset_id)
            )
            self.assertEqual(len(versions), 1)
            self.assertEqual(versions[0].version, "1")

        asyncio.run(async_test())

    def test_get_non_existent_specific_version(self):
        async def async_test():
            asset_type = "shape"
            asset_id = f"asset_{uuid4()}"
            await self.store.create(asset_type, asset_id, b"data_v1")

            non_existent_version_uri = AssetUri.build(
                asset_type=asset_type,
                asset_id=asset_id,
                version="99",  # Assuming version 99 won't exist
            )
            with self.assertRaises(FileNotFoundError):
                await self.store.get(non_existent_version_uri)

        asyncio.run(async_test())

    def test_delete_last_version(self):
        async def async_test():
            asset_type = "shape"
            asset_id = f"asset_{uuid4()}"

            uri_v1 = await self.store.create(asset_type, asset_id, b"v1_data")
            uri_v2 = await self.store.update(uri_v1, b"v2_data")

            await self.store.delete(uri_v2)  # Delete latest version
            with self.assertRaises(FileNotFoundError):
                await self.store.get(uri_v2)

            # v1 should still exist
            self.assertEqual(await self.store.get(uri_v1), b"v1_data")
            versions_after_v2_delete = await self.store.list_versions(uri_v1)
            self.assertEqual(len(versions_after_v2_delete), 1)
            self.assertEqual(versions_after_v2_delete[0].version, "1")

            await self.store.delete(uri_v1)  # Delete the now last version (v1)
            with self.assertRaises(FileNotFoundError):
                await self.store.get(uri_v1)

            versions_after_all_delete = await self.store.list_versions(uri_v1)
            self.assertEqual(len(versions_after_all_delete), 0)

            # Asset should not appear in list_assets
            listed_assets = await self.store.list_assets(asset_type=asset_type)
            self.assertFalse(any(a.asset_id == asset_id for a in listed_assets))
            self.assertTrue(await self.store.is_empty(asset_type))

        asyncio.run(async_test())

    def test_get_latest_on_non_existent_asset(self):
        async def async_test():
            latest_uri = AssetUri.build(
                asset_type="shape",
                asset_id=f"non_existent_id_for_latest_{uuid4()}",
                version="latest",
            )
            with self.assertRaises(
                FileNotFoundError
            ):  # Or custom NoVersionsFoundError if that's how store behaves
                await self.store.get(latest_uri)

        asyncio.run(async_test())


class TestPathToolFileStore(BaseTestPathToolAssetStore):
    """Test suite for FileStore with full versioning support."""

    def setUp(self):
        super().setUp()
        asset_type_map = {
            "*": "{asset_type}/{asset_id}/{version}",
            "special1": "Especial/{asset_id}/{version}",
            "special2": "my/super/{asset_id}.spcl",
        }
        self.store = FileStore("versioned", self.tmp_path, asset_type_map)

    def test_get_latest_version(self):
        async def async_test():
            asset_type = f"latest_{uuid4()}"
            asset_id = f"asset_{uuid4()}"

            uri1 = await self.store.create(asset_type, asset_id, b"v1")
            await self.store.update(uri1, b"v2")

            latest_uri = AssetUri.build(asset_type=asset_type, asset_id=asset_id, version="latest")
            self.assertEqual(await self.store.get(latest_uri), b"v2")

        asyncio.run(async_test())

    def test_delete_all_versions(self):
        async def async_test():
            asset_type = f"delete_{uuid4()}"
            asset_id = f"asset_{uuid4()}"

            uri1 = await self.store.create(asset_type, asset_id, b"v1")
            await self.store.update(uri1, b"v2")

            uri = AssetUri.build(asset_type=asset_type, asset_id=asset_id)
            await self.store.delete(uri)

            with self.assertRaises(FileNotFoundError):
                await self.store.get(uri1)

        asyncio.run(async_test())


class TestPathToolMemoryStore(BaseTestPathToolAssetStore):
    """Test suite for MemoryStore."""

    def setUp(self):
        super().setUp()
        self.store = MemoryStore("memory_test")


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2013 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

# Unit tests for the Arch wall module

import os
import Arch
import Draft
import Part
import FreeCAD as App
from bimtests import TestArchBase


class TestArchWall(TestArchBase.TestArchBase):

    def testWall(self):
        operation = "Checking Arch Wall..."
        self.printTestMessage(operation)

        l = Draft.makeLine(App.Vector(0, 0, 0), App.Vector(-2, 0, 0))
        w = Arch.makeWall(l)
        self.assertTrue(w, "Arch Wall failed")

    def testWallMultiMatAlign(self):
        operation = "Checking Arch Wall with MultiMaterial and 3 alignments..."
        self.printTestMessage(operation)

        matA = Arch.makeMaterial()
        matB = Arch.makeMaterial()
        matMulti = Arch.makeMultiMaterial()
        matMulti.Materials = [matA, matB]
        matMulti.Thicknesses = [100, 200]  # total width different from default 200
        pts = [
            App.Vector(0, 0, 0),
            App.Vector(1000, 0, 0),
            App.Vector(1000, 1000, 0),
            App.Vector(2000, 1000, 0),
        ]
        # wall based on wire:
        wire = Draft.makeWire(pts)
        wallWire = Arch.makeWall(wire)
        wallWire.Material = matMulti
        # wall based on sketch:
        sketch = App.activeDocument().addObject("Sketcher::SketchObject", "Sketch")
        sketch.addGeometry(
            [
                Part.LineSegment(pts[0], pts[1]),
                Part.LineSegment(pts[1], pts[2]),
                Part.LineSegment(pts[2], pts[3]),
            ]
        )
        wallSketch = Arch.makeWall(sketch)
        wallSketch.Material = matMulti

        alignLst = ["Left", "Center", "Right"]
        checkLst = [
            [App.Vector(0, -300, 0), App.Vector(2000, 1000, 0)],
            [App.Vector(0, -150, 0), App.Vector(2000, 1150, 0)],
            [App.Vector(0, 0, 0), App.Vector(2000, 1300, 0)],
        ]
        for i in range(3):
            wallWire.Align = alignLst[i]
            wallSketch.Align = alignLst[i]
            App.ActiveDocument.recompute()
            for box in [wallWire.Shape.BoundBox, wallSketch.Shape.BoundBox]:
                ptMin = App.Vector(box.XMin, box.YMin, 0)
                self.assertTrue(
                    ptMin.isEqual(checkLst[i][0], 1e-8),
                    "Arch Wall with MultiMaterial and 3 alignments failed",
                )
                ptMax = App.Vector(box.XMax, box.YMax, 0)
                self.assertTrue(
                    ptMax.isEqual(checkLst[i][1], 1e-8),
                    "Arch Wall with MultiMaterial and 3 alignments failed",
                )

    def test_makeWall(self):
        """Test the makeWall function."""
        operation = "Testing makeWall function"
        self.printTestMessage(operation)

        wall = Arch.makeWall(length=5000, width=200, height=3000)
        self.assertIsNotNone(wall, "makeWall failed to create a wall object.")
        self.assertEqual(wall.Label, "Wall", "Wall label is incorrect.")

    def test_joinWalls(self):
        """Test the joinWalls function."""
        operation = "Testing joinWalls function"
        self.printTestMessage(operation)

        base_line1 = Draft.makeLine(App.Vector(0, 0, 0), App.Vector(5000, 0, 0))
        base_line2 = Draft.makeLine(App.Vector(5000, 0, 0), App.Vector(5000, 3000, 0))
        wall1 = Arch.makeWall(base_line1, width=200, height=3000)
        wall2 = Arch.makeWall(base_line2, width=200, height=3000)
        joined_wall = Arch.joinWalls([wall1, wall2])
        self.assertIsNotNone(joined_wall, "joinWalls failed to join walls.")

    def test_remove_base_from_wall_without_host(self):
        """
        Tests that removing a wall's base using removeComponents(host=None)
        does not crash and successfully unlinks the base.
        This is the non-regression test for the 'list' has no attribute 'Base' bug.
        https://github.com/FreeCAD/FreeCAD/issues/24532
        """
        self.printTestMessage("Testing removal of a wall's base component...")

        # 1. Arrange: Create a wall with a base
        line = Draft.makeLine(App.Vector(0, 0, 0), App.Vector(2000, 0, 0))
        wall = Arch.makeWall(line)
        self.assertIsNotNone(wall.Base, "Pre-condition failed: Wall should have a base.")

        # 2. Act: Call removeComponents on the base, simulating the failing workflow
        # Before the fix, this will raise an AttributeError.
        # After the fix, it should complete without error.
        Arch.removeComponents([wall.Base])
        self.document.recompute()

        # 3. Assert: The base should now be None
        self.assertIsNone(wall.Base, "The wall's Base property was not cleared after removal.")



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2025 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

import re
import os
from typing import Optional

from PySideWrapper import QtCore, QtGui, QtSvg

try:
    # If this system provides a secure parser, use that:
    import defusedxml.ElementTree as ET
except ImportError:
    # Otherwise fall back to the Python standard parser
    import xml.etree.ElementTree as ET

from Addon import Addon
import addonmanager_freecad_interface as fci

MAX_ICON_BYTES = 10 * 1024 * 1024

SVG_ROOT_RE = re.compile(
    rb"""^\s*(?:\xEF\xBB\xBF)?(?:<!--.*?-->|\s|<\?xml[^>]*\?>|<!DOCTYPE[^>]*>)*<\s*svg(?=[\s>])""",
    re.IGNORECASE | re.DOTALL | re.VERBOSE,
)


def icon_from_bytes(raw: bytes) -> QtGui.QIcon:
    """Given raw bytes, try to create the best icon we can. If given SVG (either raw or compressed),
    this will result in a QIcon backed by a scalable QIconEngine. Otherwise, it's just a bitmap."""
    if is_svg_bytes(raw):
        return scalable_icon_from_svg_bytes(raw)
    elif is_gzip(raw):
        decompressed = decompress_gzip_limited(raw)
        if decompressed is not None:
            return scalable_icon_from_svg_bytes(raw)  # Qt will handle the compressed data for us
    icon = QtGui.QIcon(QtGui.QPixmap.fromImage(QtGui.QImage.fromData(raw)))
    if icon.isNull():
        raise BadIconData("Icon data is not in a recognized image file format")
    return icon


def is_valid_xml(svg_bytes: bytes) -> bool:
    """Returns True if the given SVG bytes are at least a valid XML file, False otherwise."""
    try:
        _ = ET.fromstring(svg_bytes.decode("utf-8"))
    except ET.ParseError:
        return False
    except UnicodeDecodeError:
        return False
    except RuntimeError:
        return False
    return True


class BadIconData(Exception):
    pass


def is_svg_bytes(raw: bytes) -> bool:
    head = raw[:MAX_ICON_BYTES]
    if SVG_ROOT_RE.search(head):
        if is_valid_xml(raw):
            return True
        raise BadIconData("File header looks like SVG, but data is invalid")
    return False


def is_gzip(data: bytes) -> bool:
    return len(data) >= 2 and data[0] == 0x1F and data[1] == 0x8B


MAX_GZIP_EXPANSION_RATIO = 16
MAX_GZIP_OUTPUT_ABS = 512 * 1024  # 512 KiB


def decompress_gzip_limited(data: bytes) -> Optional[bytes]:
    """Allow compressed size ≤ MAX_ICON_BYTES; read at most a small, bounded amount.
    Returns None on failure or if output would exceed the bound."""
    if not isinstance(data, (bytes, bytearray, memoryview)):
        return None
    if len(data) > MAX_ICON_BYTES:
        return None

    import io, gzip, zlib

    max_out = min(MAX_GZIP_OUTPUT_ABS, MAX_GZIP_EXPANSION_RATIO * len(data))
    try:
        with gzip.GzipFile(fileobj=io.BytesIO(data)) as f:
            out = f.read(max_out + 1)  # stream, don’t inflate unbounded
        if len(out) > max_out:
            return None
        return out
    except (OSError, EOFError, zlib.error, ValueError, TypeError):
        return None


class SvgIconEngine(QtGui.QIconEngine):
    def __init__(self, svg_bytes: bytes):
        super().__init__()
        self.renderer = QtSvg.QSvgRenderer(QtCore.QByteArray(svg_bytes))

    def paint(self, painter: QtGui.QPainter, rect: QtCore.QRect, mode, state):
        self.renderer.render(painter, rect)

    def pixmap(self, size: QtCore.QSize, mode, state):
        pixmap = QtGui.QPixmap(size)
        pixmap.fill(QtCore.Qt.transparent)  # type: ignore[arg-type]
        painter = QtGui.QPainter(pixmap)
        self.renderer.render(painter)
        painter.end()
        return pixmap


def scalable_icon_from_svg_bytes(svg_bytes: bytes) -> QtGui.QIcon:
    engine = SvgIconEngine(svg_bytes)
    return QtGui.QIcon(engine)


cached_default_icons = {}


def get_icon_for_addon(addon: Addon, update: bool = False) -> QtGui.QIcon:
    """Returns an icon for an Addon.
    :param addon: The addon to get an icon for.
    :param update: If True, the icon will be updated even if it already exists.
    :return: The QIcon for the addon. Scalable if it was SVG, otherwise a bitmap"""

    icon_path = os.path.join(os.path.dirname(__file__), "Resources", "icons")

    if not update and addon.icon and not addon.icon.isNull():
        return addon.icon
    elif addon.icon_data:
        if len(addon.icon_data) > MAX_ICON_BYTES:
            fci.Console.PrintWarning(
                f"WARNING: Icon file for addon '{addon.display_name}' is too large (max size is {MAX_ICON_BYTES} bytes)\n"
            )
        try:
            addon.icon = icon_from_bytes(addon.icon_data)
            return addon.icon
        except BadIconData as e:
            fci.Console.PrintWarning(
                f"Icon file for addon '{addon.display_name}' is invalid:\n{e}\n"
            )
    elif addon.macro:
        if addon.macro.icon_data:
            if len(addon.macro.icon_data) > MAX_ICON_BYTES:
                fci.Console.PrintWarning(
                    f"WARNING: Icon data for macro '{addon.display_name}' is too large (max size is {MAX_ICON_BYTES} bytes)\n"
                )
            try:
                addon.icon = icon_from_bytes(addon.macro.icon_data)
                return addon.icon
            except BadIconData as e:
                fci.Console.PrintWarning(
                    f"Icon data for macro '{addon.display_name}' is invalid:\n{e}\n"
                )
        elif addon.macro.xpm:
            xpm = QtGui.QImage.fromData(addon.macro.xpm.strip().encode("utf-8"), format="XPM")  # type: ignore[arg-type]
            if xpm.isNull() or xpm.width() == 0 or xpm.height() == 0:
                fci.Console.PrintWarning(
                    f"The XPM icon data for macro '{addon.display_name}' is invalid (please report this to the macro's author, {addon.macro.author})\n"
                )
            else:
                addon.icon = QtGui.QIcon(QtGui.QPixmap.fromImage(xpm))
                return addon.icon

    if not cached_default_icons:
        cached_default_icons["package"] = QtGui.QIcon(
            os.path.join(icon_path, "document-package.svg")
        )
        cached_default_icons["macro"] = QtGui.QIcon(os.path.join(icon_path, "document-python.svg"))
        cached_default_icons["workbench"] = QtGui.QIcon(
            os.path.join(icon_path, "document-package.svg")
        )

    if addon.repo_type == Addon.Kind.WORKBENCH:
        return cached_default_icons["package"]
    if addon.repo_type == Addon.Kind.MACRO:
        return cached_default_icons["macro"]

    return cached_default_icons["package"]



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2019 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the viewprovider code for the Wire (polyline) object.

This viewprovider is also used by simple lines, B-splines, bezier curves,
and similar objects.
"""
## @package view_wire
# \ingroup draftviewproviders
# \brief Provides the viewprovider code for the Wire (polyline) object.

## \addtogroup draftviewproviders
# @{
import pivy.coin as coin
import PySide.QtCore as QtCore
import PySide.QtGui as QtGui
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import FreeCADGui as Gui
import DraftVecUtils
import WorkingPlane
from draftgeoutils import wires
from draftutils import gui_utils
from draftutils import params
from draftutils import utils
from draftutils.translate import translate
from draftviewproviders.view_base import ViewProviderDraft


class ViewProviderWire(ViewProviderDraft):
    """A base View Provider for the Wire object."""

    def __init__(self, vobj):
        super(ViewProviderWire, self).__init__(vobj)
        self._set_properties(vobj)

    def _set_properties(self, vobj):
        """Set the properties of objects if they don't exist."""
        super(ViewProviderWire, self)._set_properties(vobj)

        if not hasattr(vobj, "ArrowSizeStart"):
            _tip = QT_TRANSLATE_NOOP("App::Property", "Arrow size")
            vobj.addProperty("App::PropertyLength", "ArrowSizeStart", "Draft", _tip, locked=True)
            vobj.ArrowSizeStart = params.get_param("arrowsizestart")

        if not hasattr(vobj, "ArrowTypeStart"):
            _tip = QT_TRANSLATE_NOOP("App::Property", "Arrow type")
            vobj.addProperty(
                "App::PropertyEnumeration", "ArrowTypeStart", "Draft", _tip, locked=True
            )
            vobj.ArrowTypeStart = utils.ARROW_TYPES
            vobj.ArrowTypeStart = "None"

        if not hasattr(vobj, "ArrowSizeEnd"):
            _tip = QT_TRANSLATE_NOOP("App::Property", "Arrow size")
            vobj.addProperty("App::PropertyLength", "ArrowSizeEnd", "Draft", _tip, locked=True)
            vobj.ArrowSizeEnd = params.get_param("arrowsizeend")

        if not hasattr(vobj, "ArrowTypeEnd"):
            _tip = QT_TRANSLATE_NOOP("App::Property", "Arrow type")
            vobj.addProperty("App::PropertyEnumeration", "ArrowTypeEnd", "Draft", _tip, locked=True)
            vobj.ArrowTypeEnd = utils.ARROW_TYPES
            vobj.ArrowTypeEnd = "None"

    def attach(self, vobj):
        self.Object = vobj.Object

        self.color = coin.SoBaseColor()
        self.color.rgb.setValue(*vobj.LineColor[:3])

        self.startSymbol = gui_utils.dim_symbol()
        self.coords1 = coin.SoTransform()
        self.pt1 = coin.SoSeparator()
        self.pt1.addChild(self.color)
        self.pt1.addChild(self.coords1)
        self.pt1.addChild(self.startSymbol)

        self.endSymbol = gui_utils.dim_symbol()
        self.coords2 = coin.SoTransform()
        self.pt2 = coin.SoSeparator()
        self.pt2.addChild(self.color)
        self.pt2.addChild(self.coords2)
        self.pt2.addChild(self.endSymbol)

        super(ViewProviderWire, self).attach(vobj)
        self.onChanged(vobj, "ArrowSizeStart")

    def updateData(self, obj, prop):
        if (
            prop == "Points"
            and hasattr(obj, "Points")
            and len(obj.Points) >= 2
            and hasattr(self, "coords1")
            and hasattr(self, "coords2")
        ):
            self.coords1.translation.setValue(*obj.Points[0])
            v1 = obj.Points[1].sub(obj.Points[0])
            if not DraftVecUtils.isNull(v1):
                v1.normalize()
                rot1 = coin.SbRotation()
                rot1.setValue(coin.SbVec3f(1, 0, 0), coin.SbVec3f(*v1))
                self.coords1.rotation.setValue(rot1)
            self.coords2.translation.setValue(*obj.Points[-1])
            v2 = obj.Points[-2].sub(obj.Points[-1])
            if not DraftVecUtils.isNull(v2):
                v2.normalize()
                rot2 = coin.SbRotation()
                rot2.setValue(coin.SbVec3f(1, 0, 0), coin.SbVec3f(*v2))
                self.coords2.rotation.setValue(rot2)

    def onChanged(self, vobj, prop):
        if (
            prop
            in ["ArrowSizeStart", "ArrowSizeEnd", "ArrowTypeStart", "ArrowTypeEnd", "Visibility"]
            and hasattr(vobj, "ArrowSizeStart")
            and hasattr(vobj, "ArrowSizeEnd")
            and hasattr(vobj, "ArrowTypeStart")
            and hasattr(vobj, "ArrowTypeEnd")
            and hasattr(vobj, "Visibility")
            and hasattr(self, "pt1")
            and hasattr(self, "pt2")
        ):
            rn = vobj.RootNode
            rn.removeChild(self.pt1)
            rn.removeChild(self.pt2)
            if vobj.Visibility:
                self.pt1.removeChild(self.startSymbol)
                self.startSymbol = gui_utils.dim_symbol(
                    utils.ARROW_TYPES.index(vobj.ArrowTypeStart)
                )
                self.pt1.addChild(self.startSymbol)
                self.coords1.scaleFactor.setValue([vobj.ArrowSizeStart] * 3)

                self.pt2.removeChild(self.endSymbol)
                self.endSymbol = gui_utils.dim_symbol(utils.ARROW_TYPES.index(vobj.ArrowTypeEnd))
                self.pt2.addChild(self.endSymbol)
                self.coords2.scaleFactor.setValue([vobj.ArrowSizeEnd] * 3)

                self.updateData(vobj.Object, "Points")
                rn.addChild(self.pt1)
                rn.addChild(self.pt2)

        if (
            prop == "LineColor"
            and hasattr(vobj, "LineColor")
            and hasattr(self, "pt1")
            and hasattr(self, "pt2")
        ):
            self.color.rgb.setValue(*vobj.LineColor[:3])

        super().onChanged(vobj, prop)
        return

    def claimChildren(self):
        if hasattr(self.Object, "Base"):
            return [self.Object.Base, self.Object.Tool]
        return []

    def flatten(self):  # Only to be used for Draft_Wires.
        if not hasattr(self, "Object"):
            return

        wp = WorkingPlane.get_working_plane()
        flat_wire = wires.flattenWire(
            self.Object.Shape.Wires[0], origin=wp.position, normal=wp.axis
        )

        doc = App.ActiveDocument
        doc.openTransaction(translate("draft", "Flatten"))

        self.Object.Placement = App.Placement()
        self.Object.Points = [vert.Point for vert in flat_wire.Vertexes]
        self.Object.Closed = flat_wire.isClosed()

        doc.recompute()
        doc.commitTransaction()


# Alias for compatibility with v0.18 and earlier
_ViewProviderWire = ViewProviderWire

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2020 Antoine Lafr                                       *
# *   Copyright (c) 2025 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""Unit tests for the Draft Workbench, DraftGeomUtils module tests."""

import Part
import DraftGeomUtils
from FreeCAD import Vector
from drafttests import test_base
from draftutils.messages import _msg


class TestDraftGeomUtils(test_base.DraftTestCaseNoDoc):
    """Testing the functions in the file DraftGeomUtils.py"""

    def check_wire(self, wire):
        offset_values = (2000.0, 0.0, -1000, -2000, -3000, -5500)
        for offset_start in offset_values:
            for offset_end in offset_values:
                if offset_start + offset_end > -wire.Length:
                    # print ("'start={0}, end={1}' failed".format(offset_start, offset_end))
                    try:
                        extended = DraftGeomUtils.get_extended_wire(wire, offset_start, offset_end)
                        # Test that the extended wire's length is correctly changed
                        self.assertAlmostEqual(
                            extended.Length,
                            wire.Length + offset_start + offset_end,
                            DraftGeomUtils.precision(),
                            "'start={0}, end={1}' failed".format(offset_start, offset_end),
                        )
                        if offset_start == 0.0:
                            # If offset_start is 0.0, check that the wire's start point is unchanged
                            self.assertAlmostEqual(
                                extended.OrderedVertexes[0].Point.distanceToPoint(
                                    wire.OrderedVertexes[0].Point
                                ),
                                0.0,
                                DraftGeomUtils.precision(),
                                "'start={0}, end={1}' failed".format(offset_start, offset_end),
                            )
                        if offset_end == 0.0:
                            # If offset_end is 0.0, check that the wire's end point is unchanged
                            self.assertAlmostEqual(
                                extended.OrderedVertexes[-1].Point.distanceToPoint(
                                    wire.OrderedVertexes[-1].Point
                                ),
                                0.0,
                                DraftGeomUtils.precision(),
                                "'start={0}, end={1}' failed".format(offset_start, offset_end),
                            )
                    except Exception as exc:
                        print(
                            "get_extended_wire failed for 'start={0}, end={1}'".format(
                                offset_start, offset_end
                            )
                        )
                        raise exc

    def test_get_extended_wire1(self):
        """Test the DraftGeomUtils.get_extended_wire function."""
        operation = "DraftGeomUtils.get_extended_wire1"
        _msg("  Test '{}'".format(operation))

        # Build wires made with straight edges and various combination of Orientation: the wires 1-4 are all equivalent
        points = [
            Vector(0.0, 0.0, 0.0),
            Vector(1500.0, 2000.0, 0.0),
            Vector(4500.0, 2000.0, 0.0),
            Vector(4500.0, 2000.0, 2500.0),
        ]

        edges = []
        for start, end in zip(points[:-1], points[1:]):
            edge = Part.makeLine(start, end)
            edges.append(edge)
        wire = Part.Wire(edges)
        self.check_wire(wire)

    def test_get_extended_wire2(self):
        """Test the DraftGeomUtils.get_extended_wire function."""
        operation = "DraftGeomUtils.get_extended_wire2"
        _msg("  Test '{}'".format(operation))

        # Build wires made with straight edges and various combination of Orientation: the wires 1-4 are all equivalent
        points = [
            Vector(0.0, 0.0, 0.0),
            Vector(1500.0, 2000.0, 0.0),
            Vector(4500.0, 2000.0, 0.0),
            Vector(4500.0, 2000.0, 2500.0),
        ]

        edges = []
        for start, end in zip(points[:-1], points[1:]):
            edge = Part.makeLine(end, start)
            edge.Orientation = "Reversed"
            edges.append(edge)
        wire = Part.Wire(edges)
        self.check_wire(wire)

    def test_get_extended_wire3(self):
        """Test the DraftGeomUtils.get_extended_wire function."""
        operation = "DraftGeomUtils.get_extended_wire3"
        _msg("  Test '{}'".format(operation))

        # Build wires made with straight edges and various combination of Orientation: the wires 1-4 are all equivalent
        points = [
            Vector(0.0, 0.0, 0.0),
            Vector(1500.0, 2000.0, 0.0),
            Vector(4500.0, 2000.0, 0.0),
            Vector(4500.0, 2000.0, 2500.0),
        ]

        edges = []
        for start, end in zip(points[:-1], points[1:]):
            edge = Part.makeLine(start, end)
            edge.Orientation = "Reversed"
            edges.insert(0, edge)
        wire = Part.Wire(edges)
        wire.Orientation = "Reversed"
        self.check_wire(wire)

    def test_get_extended_wire4(self):
        """Test the DraftGeomUtils.get_extended_wire function."""
        operation = "DraftGeomUtils.get_extended_wire4"
        _msg("  Test '{}'".format(operation))

        # Build wires made with straight edges and various combination of Orientation: the wires 1-4 are all equivalent
        points = [
            Vector(0.0, 0.0, 0.0),
            Vector(1500.0, 2000.0, 0.0),
            Vector(4500.0, 2000.0, 0.0),
            Vector(4500.0, 2000.0, 2500.0),
        ]

        edges = []
        for start, end in zip(points[:-1], points[1:]):
            edge = Part.makeLine(end, start)
            edges.insert(0, edge)
        wire = Part.Wire(edges)
        wire.Orientation = "Reversed"
        self.check_wire(wire)

    def test_get_extended_wire5(self):
        """Test the DraftGeomUtils.get_extended_wire function."""
        operation = "DraftGeomUtils.get_extended_wire5"
        _msg("  Test '{}'".format(operation))

        # Build wires made with arcs and various combination of Orientation: the wires 5-8 are all equivalent
        points = [
            Vector(0.0, 0.0, 0.0),
            Vector(1000.0, 1000.0, 0.0),
            Vector(2000.0, 0.0, 0.0),
            Vector(3000.0, 0.0, 1000.0),
            Vector(4000.0, 0.0, 0.0),
        ]

        edges = []
        for start, mid, end in zip(points[:-2], points[1:-1], points[2:]):
            edge = Part.Arc(start, mid, end).toShape()
            edges.append(edge)
        wire = Part.Wire(edges)
        self.check_wire(wire)

    def test_get_extended_wire6(self):
        """Test the DraftGeomUtils.get_extended_wire function."""
        operation = "DraftGeomUtils.get_extended_wire6"
        _msg("  Test '{}'".format(operation))

        # Build wires made with arcs and various combination of Orientation: the wires 5-8 are all equivalent
        points = [
            Vector(0.0, 0.0, 0.0),
            Vector(1000.0, 1000.0, 0.0),
            Vector(2000.0, 0.0, 0.0),
            Vector(3000.0, 0.0, 1000.0),
            Vector(4000.0, 0.0, 0.0),
        ]

        edges = []
        for start, mid, end in zip(points[:-2], points[1:-1], points[2:]):
            edge = Part.Arc(end, mid, start).toShape()
            edge.Orientation = "Reversed"
            edges.append(edge)
        wire = Part.Wire(edges)
        self.check_wire(wire)

    def test_get_extended_wire7(self):
        """Test the DraftGeomUtils.get_extended_wire function."""
        operation = "DraftGeomUtils.get_extended_wire7"
        _msg("  Test '{}'".format(operation))

        # Build wires made with arcs and various combination of Orientation: the wires 5-8 are all equivalent
        points = [
            Vector(0.0, 0.0, 0.0),
            Vector(1000.0, 1000.0, 0.0),
            Vector(2000.0, 0.0, 0.0),
            Vector(3000.0, 0.0, 1000.0),
            Vector(4000.0, 0.0, 0.0),
        ]

        edges = []
        for start, mid, end in zip(points[:-2], points[1:-1], points[2:]):
            edge = Part.Arc(start, mid, end).toShape()
            edge.Orientation = "Reversed"
            edges.insert(0, edge)
        wire = Part.Wire(edges)
        wire.Orientation = "Reversed"
        self.check_wire(wire)

    def test_get_extended_wire8(self):
        """Test the DraftGeomUtils.get_extended_wire function."""
        operation = "DraftGeomUtils.get_extended_wire8"
        _msg("  Test '{}'".format(operation))

        # Build wires made with arcs and various combination of Orientation: the wires 5-8 are all equivalent
        points = [
            Vector(0.0, 0.0, 0.0),
            Vector(1000.0, 1000.0, 0.0),
            Vector(2000.0, 0.0, 0.0),
            Vector(3000.0, 0.0, 1000.0),
            Vector(4000.0, 0.0, 0.0),
        ]

        edges = []
        for start, mid, end in zip(points[:-2], points[1:-1], points[2:]):
            edge = Part.Arc(end, mid, start).toShape()
            edges.insert(0, edge)
        wire = Part.Wire(edges)
        wire.Orientation = "Reversed"
        self.check_wire(wire)


# suite = unittest.defaultTestLoader.loadTestsFromTestCase(TestDraftGeomUtils)
# unittest.TextTestRunner().run(suite)



# SPDX-License-Identifier: LGPL-2.1-or-later

################################################################################
#                                                                              #
#   © 2026 Billy Huddleston <billy@ivdc.com>                                   #
#                                                                              #
#   FreeCAD is free software: you can redistribute it and/or modify            #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   FreeCAD is distributed in the hope that it will be useful,                 #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with FreeCAD. If not, see https://www.gnu.org/licenses       #
#                                                                              #
################################################################################

import FreeCAD
from typing import Tuple, Mapping
from .base import ToolBitShape


class ToolBitShapeTaperedBallNose(ToolBitShape):
    name: str = "TaperedBallNose"
    aliases = ("taperedballnose",)

    @classmethod
    def schema(cls) -> Mapping[str, Tuple[str, str]]:
        return {
            "CuttingEdgeHeight": (
                FreeCAD.Qt.translate("ToolBitShape", "Cutting edge height"),
                "App::PropertyLength",
            ),
            "Diameter": (
                FreeCAD.Qt.translate("ToolBitShape", "Diameter"),
                "App::PropertyLength",
            ),
            "Flutes": (
                FreeCAD.Qt.translate("ToolBitShape", "Flutes"),
                "App::PropertyInteger",
            ),
            "Length": (
                FreeCAD.Qt.translate("ToolBitShape", "Overall tool length"),
                "App::PropertyLength",
            ),
            "ShankDiameter": (
                FreeCAD.Qt.translate("ToolBitShape", "Shank diameter"),
                "App::PropertyLength",
            ),
            "TaperAngle": (
                FreeCAD.Qt.translate("ToolBitShape", "Included Taper angle"),
                "App::PropertyAngle",
            ),
            "TaperDiameter": (
                FreeCAD.Qt.translate("ToolBitShape", "Diameter at top of Taper"),
                "App::PropertyLength",
            ),
        }

    @property
    def label(self) -> str:
        return FreeCAD.Qt.translate("ToolBitShape", "Tapered Ball Nose")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 FreeCAD Developers                                 *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to create Polygon objects."""
## @package make_polygon
# \ingroup draftmake
# \brief Provides functions to create Polygon objects.

## \addtogroup draftmake
# @{
import FreeCAD as App
import draftutils.gui_utils as gui_utils

from draftobjects.polygon import Polygon
from draftviewproviders.view_base import ViewProviderDraft


def make_polygon(nfaces, radius=1, inscribed=True, placement=None, face=None, support=None):
    """makePolgon(edges,[radius],[inscribed],[placement],[face])

    Creates a polygon object with the given number of edges and radius.

    Parameters
    ----------
    edges : int
        Number of edges of the polygon.

    radius :
        Radius of the control circle.

    inscribed : bool
        Defines is the polygon is inscribed or not into the control circle.

    placement : Base.Placement
        If placement is given, it is used.

    face : bool
        If face is True, the resulting shape is displayed as a face,
        otherwise as a wireframe.

    support :
        TODO: Describe
    """
    if not App.ActiveDocument:
        App.Console.PrintError("No active document. Aborting\n")
        return
    if nfaces < 3:
        return None
    obj = App.ActiveDocument.addObject("Part::Part2DObjectPython", "Polygon")
    Polygon(obj)
    obj.FacesNumber = nfaces
    obj.Radius = radius
    if face is not None:
        obj.MakeFace = face
    if inscribed:
        obj.DrawMode = "inscribed"
    else:
        obj.DrawMode = "circumscribed"
    obj.AttachmentSupport = support
    if placement:
        obj.Placement = placement
    if App.GuiUp:
        ViewProviderDraft(obj.ViewObject)
        gui_utils.format_object(obj)
        gui_utils.select(obj)

    return obj


makePolygon = make_polygon

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
from typing import Optional
import FreeCADGui
from functools import partial
from PySide import QtGui
from ...camassets import cam_assets
from .. import ToolBitShape
from .flowlayout import FlowLayout
from .shapebutton import ShapeButton


class ShapeSelector:
    def __init__(self):
        self.shape = None
        self.form = FreeCADGui.PySideUic.loadUi(":/panels/ShapeSelector.ui")

        self.form.buttonBox.clicked.connect(self.form.close)

        self.flows = {}

        self.update_shapes()

    def _add_shape_group(self, toolbox):
        if toolbox in self.flows:
            return self.flows[toolbox]
        flow = FlowLayout(toolbox, orientation=QtGui.Qt.Horizontal)
        flow.widthChanged.connect(lambda x: toolbox.setMinimumWidth(x))
        self.flows[toolbox] = flow
        return flow

    def _add_shapes(self, toolbox, shapes):
        flow = self._add_shape_group(toolbox)

        # Remove all shapes first.
        for i in reversed(range(flow.count())):
            flow.itemAt(i).widget().setParent(None)

        # Add all shapes.
        for shape in sorted(shapes, key=lambda x: x.label):
            button = ShapeButton(shape)
            flow.addWidget(button)
            cb = partial(self.on_shape_button_clicked, shape)
            button.clicked.connect(cb)

    def update_shapes(self):
        # Retrieve each shape asset
        builtin = cam_assets.fetch(asset_type="toolbitshape", store="builtin")
        builtin = {c.id: c for c in builtin}
        custom = cam_assets.fetch(asset_type="toolbitshape", store="local")
        for shape in custom:
            builtin.pop(shape.id, None)

        # Combine all shapes into a single list
        all_shapes = list(builtin.values()) + list(custom)
        self._add_shapes(self.form.toolsContainer, all_shapes)

    def on_shape_button_clicked(self, shape):
        self.shape = shape
        self.form.close()

    def show(self) -> Optional[ToolBitShape]:
        self.form.exec()
        return self.shape



# SPDX-License-Identifier: LGPL-2.1-or-later
# /**************************************************************************
#                                                                           *
#    Copyright (c) 2024 Ondsel <development@ondsel.com>                     *
#                                                                           *
#    This file is part of FreeCAD.                                          *
#                                                                           *
#    FreeCAD is free software: you can redistribute it and/or modify it     *
#    under the terms of the GNU Lesser General Public License as            *
#    published by the Free Software Foundation, either version 2.1 of the   *
#    License, or (at your option) any later version.                        *
#                                                                           *
#    FreeCAD is distributed in the hope that it will be useful, but         *
#    WITHOUT ANY WARRANTY; without even the implied warranty of             *
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
#    Lesser General Public License for more details.                        *
#                                                                           *
#    You should have received a copy of the GNU Lesser General Public       *
#    License along with FreeCAD. If not, see                                *
#    <https://www.gnu.org/licenses/>.                                       *
#                                                                           *
# **************************************************************************/

import re
import os
import time
import FreeCAD as App

from pivy import coin
from Part import LineSegment, Compound

from PySide.QtCore import QT_TRANSLATE_NOOP

if App.GuiUp:
    import FreeCADGui as Gui
    from PySide import QtCore, QtGui, QtWidgets
    from PySide.QtWidgets import (
        QPushButton,
        QMenu,
        QDialog,
        QComboBox,
        QLineEdit,
        QGridLayout,
        QLabel,
        QDialogButtonBox,
    )
    from PySide.QtCore import Qt, QPoint
    from PySide.QtGui import QCursor, QIcon, QGuiApplication

import UtilsAssembly
import Preferences

translate = App.Qt.translate

__title__ = "Assembly Command Create Simulation"
__author__ = "Ondsel"
__url__ = "https://www.freecad.org"


class CommandCreateSimulation:
    def __init__(self):
        pass

    def GetResources(self):
        return {
            "Pixmap": "Assembly_CreateSimulation",
            "MenuText": QT_TRANSLATE_NOOP("Assembly_CreateSimulation", "Simulation"),
            "Accel": "V",
            "ToolTip": QT_TRANSLATE_NOOP(
                "Assembly_CreateSimulation",
                "Creates a new simulation of the current assembly",
            ),
            "CmdType": "ForEdit",
        }

    def IsActive(self):
        if not UtilsAssembly.isAssemblyCommandActive():
            return False

        assembly = UtilsAssembly.activeAssembly()
        joint_types = ["Revolute", "Slider", "Cylindrical"]
        joints = UtilsAssembly.getJointsOfType(assembly, joint_types)
        return len(joints) > 0

    def Activated(self):
        assembly = UtilsAssembly.activeAssembly()
        if not assembly:
            return

        self.panel = TaskAssemblyCreateSimulation()
        dialog = Gui.Control.showDialog(self.panel)
        if dialog is not None:
            dialog.setAutoCloseOnDeletedDocument(True)
            dialog.setDocumentName(App.ActiveDocument.Name)


######### Simulation Object ###########
class Simulation:
    def __init__(self, feaPy):
        feaPy.Proxy = self
        feaPy.addExtension("App::GroupExtensionPython")

        if not hasattr(feaPy, "aTimeStart"):
            feaPy.addProperty(
                "App::PropertyTime",
                "aTimeStart",
                "Simulation",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "Simulation start time.",
                ),
                locked=True,
            )

        if not hasattr(feaPy, "bTimeEnd"):
            feaPy.addProperty(
                "App::PropertyTime",
                "bTimeEnd",
                "Simulation",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "Simulation end time.",
                ),
                locked=True,
            )

        if not hasattr(feaPy, "cTimeStepOutput"):
            feaPy.addProperty(
                "App::PropertyTime",
                "cTimeStepOutput",
                "Simulation",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "Simulation time step for output.",
                ),
                locked=True,
            )

        if not hasattr(feaPy, "fGlobalErrorTolerance"):
            feaPy.addProperty(
                "App::PropertyFloat",
                "fGlobalErrorTolerance",
                "Simulation",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "Integration global error tolerance.",
                ),
                locked=True,
            )

        if not hasattr(feaPy, "jFramesPerSecond"):
            feaPy.addProperty(
                "App::PropertyInteger",
                "jFramesPerSecond",
                "Simulation",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "Frames Per Second.",
                ),
                locked=True,
            )

        feaPy.aTimeStart = 0.0
        feaPy.bTimeEnd = 1.0
        feaPy.cTimeStepOutput = 1.0e-2
        feaPy.fGlobalErrorTolerance = 1.0e-6
        feaPy.jFramesPerSecond = 30

        self.motionsChangedCallback = None

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def onChanged(self, feaPy, prop):
        if prop == "Group" and hasattr(self, "motionsChangedCallback"):
            if self.motionsChangedCallback is not None:
                self.motionsChangedCallback()

    def setMotionsChangedCallback(self, callback):
        self.motionsChangedCallback = callback

    def execute(self, feaPy):
        """Do something when doing a recomputation, this method is mandatory"""
        pass

    def getAssembly(self, feaPy):
        assert feaPy.isDerivedFrom("App::FeaturePython"), "Type error"
        for obj in feaPy.InList:
            if obj.isDerivedFrom("Assembly::AssemblyObject"):
                return obj
        return None


class ViewProviderSimulation:
    def __init__(self, vpDoc):
        vpDoc.Proxy = self
        self.Object = vpDoc.Object
        self.setProperties(vpDoc)

    def setProperties(self, vpDoc):
        if not hasattr(vpDoc, "Decimals"):
            vpDoc.addProperty(
                "App::PropertyInteger",
                "Decimals",
                "Space",
                QT_TRANSLATE_NOOP(
                    "App::Property", "The number of decimals to use for calculated texts"
                ),
                locked=True,
            )
            vpDoc.Decimals = 9

    def attach(self, vpDoc):
        """Setup the scene sub-graph of the view provider, this method is mandatory"""
        self.app_obj = vpDoc.Object

        self.display_mode = coin.SoType.fromName("SoFCSelection").createInstance()

        vpDoc.addDisplayMode(self.display_mode, "Wireframe")

    def updateData(self, feaPy, prop):
        """If a property of the handled feature has changed we have the chance to handle this here"""
        pass

    def getDisplayModes(self, vpDoc):
        """Return a list of display modes."""
        return ["Wireframe"]

    def getDefaultDisplayMode(self):
        """Return the name of the default display mode. It must be defined in getDisplayModes."""
        return "Wireframe"

    def onChanged(self, vpDoc, prop):
        """Here we can do something when a single property got changed"""
        pass

    def getIcon(self):
        return ":/icons/Assembly_CreateSimulation.svg"

    def dumps(self):
        """When saving the document this object gets stored using Python's json module.\
                Since we have some un-serializable parts here -- the Coin stuff -- we must define this method\
                to return a tuple of all serializable objects or None."""
        return None

    def loads(self, state):
        """When restoring the serialized object from document we have the chance to set some internals here.\
                Since no data were serialized nothing needs to be done here."""
        return None

    def claimChildren(self):
        return self.app_obj.Group

    def doubleClicked(self, vpDoc):
        task = Gui.Control.activeTaskDialog()
        if task:
            task.reject()

        assembly = vpDoc.Object.Proxy.getAssembly(vpDoc.Object)

        if assembly is None:
            return False

        if UtilsAssembly.activeAssembly() != assembly:
            Gui.ActiveDocument.setEdit(assembly)

        panel = TaskAssemblyCreateSimulation(vpDoc.Object)
        dialog = Gui.Control.showDialog(panel)
        if dialog is not None:
            dialog.setAutoCloseOnDeletedDocument(True)
            dialog.setDocumentName(App.ActiveDocument.Name)

        return True

    def onDelete(self, vobj, subelements):
        for obj in self.claimChildren():
            obj.Document.removeObject(obj.Name)
        return True


########### Motion Object #############
MotionTypes = [
    "Angular",
    "Linear",
]


class Motion:
    def __init__(self, feaPy, motionType=MotionTypes[0], joint=None, formula=""):
        feaPy.Proxy = self

        self.createProperties(feaPy)

        feaPy.MotionType = MotionTypes  # sets the list
        feaPy.MotionType = motionType  # set the initial value
        feaPy.Joint = joint
        feaPy.Formula = formula

    def onDocumentRestored(self, feaPy):
        self.createProperties(feaPy)

    def createProperties(self, feaPy):
        if not hasattr(feaPy, "Joint"):
            feaPy.addProperty(
                "App::PropertyXLinkSubHidden",
                "Joint",
                "Motion",
                QT_TRANSLATE_NOOP("App::Property", "The joint that is moved by the motion"),
                locked=True,
            )

        if not hasattr(feaPy, "Formula"):
            feaPy.addProperty(
                "App::PropertyString",
                "Formula",
                "Motion",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "This is the formula of the motion. For example '1.0*time'.",
                ),
                locked=True,
            )

        if not hasattr(feaPy, "MotionType"):
            feaPy.addProperty(
                "App::PropertyEnumeration",
                "MotionType",
                "Motion",
                QT_TRANSLATE_NOOP("App::Property", "The type of the motion"),
                locked=True,
            )

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def onChanged(self, feaPy, prop):
        pass

    def execute(self, feaPy):
        """Do something when doing a recomputation, this method is mandatory"""
        pass

    def getSimulation(self, feaPy):
        for obj in feaPy.InList:
            if hasattr(obj, "Proxy"):
                if hasattr(obj.Proxy, "setMotionsChangedCallback"):
                    return obj
        return None

    def getAssembly(self, feaPy):
        simulation = self.getSimulation(feaPy)
        if simulation is not None:
            return simulation.Proxy.getAssembly(simulation)
        return None


class ViewProviderMotion:
    def __init__(self, vp):
        vp.Proxy = self
        self.updateLabel()

    def attach(self, vpDoc):
        """Setup the scene sub-graph of the view provider, this method is mandatory"""
        self.app_obj = vpDoc.Object

        self.display_mode = coin.SoType.fromName("SoFCSelection").createInstance()

        vpDoc.addDisplayMode(self.display_mode, "Wireframe")

    def updateData(self, feaPy, prop):
        """If a property of the handled feature has changed we have the chance to handle this here"""
        pass

    def getDisplayModes(self, vpDoc):
        """Return a list of display modes."""
        return ["Wireframe"]

    def getDefaultDisplayMode(self):
        """Return the name of the default display mode. It must be defined in getDisplayModes."""
        return "Wireframe"

    def onChanged(self, vpDoc, prop):
        """Here we can do something when a single property got changed"""
        # App.Console.PrintMessage("Change property: " + str(prop) + "\n")
        pass

    def getIcon(self):
        if self.app_obj.MotionType == "Angular":
            return ":/icons/button_rotate.svg"

        return ":/icons/button_right.svg"

    def dumps(self):
        """When saving the document this object gets stored using Python's json module.\
                Since we have some un-serializable parts here -- the Coin stuff -- we must define this method\
                to return a tuple of all serializable objects or None."""
        return None

    def loads(self, state):
        """When restoring the serialized object from document we have the chance to set some internals here.\
                Since no data were serialized nothing needs to be done here."""
        return None

    def doubleClicked(self, vpDoc):
        self.openEditDialog()

    def openEditDialog(self):
        assembly = self.getAssembly()

        if assembly is None:
            return False

        joint = None
        if self.app_obj.Joint is not None:
            joint = self.app_obj.Joint[0]

        dialog = MotionEditDialog(assembly, self.app_obj.MotionType, joint, self.app_obj.Formula)
        if dialog.exec_():
            self.app_obj.MotionType = dialog.motionType
            self.app_obj.Joint = dialog.joint
            self.app_obj.Formula = dialog.formula

            self.updateLabel()

    def updateLabel(self):
        if self.app_obj.Joint is None:
            return

        typeStr = "Linear" if self.app_obj.MotionType == "Linear" else "Angular"

        self.app_obj.Label = "{label} ({type_})".format(
            label=self.app_obj.Joint[0].Label, type_=translate("Assembly", typeStr)
        )

    def getAssembly(self):
        assembly = self.app_obj.Proxy.getAssembly(self.app_obj)

        if assembly is None:
            return None

        if UtilsAssembly.activeAssembly() != assembly:
            Gui.ActiveDocument.setEdit(assembly)

        return assembly


class MotionEditDialog:
    def __init__(self, assembly, motionType=MotionTypes[0], joint=None, formula="5*time"):
        self.assembly = assembly
        self.motionType = motionType
        self.joint = joint
        self.formula = formula

        # Create a non-modal, frameless dialog
        self.dialog = QDialog()
        self.dialog.setWindowFlags(Qt.Popup)
        self.initialPos = QCursor.pos()
        self.dialog.setMinimumSize(500, 200)  # Set a reasonable minimum size

        # Create the joints combobox
        self.joint_combo = QComboBox(self.dialog)
        self.setup_joint_combo()

        # Create the motion type combobox
        self.motion_type_combo = QComboBox(self.dialog)
        self.setup_motiontype_combo()

        def on_motion_type_changed(text):
            self.motionType = text

        self.motion_type_combo.currentTextChanged.connect(on_motion_type_changed)

        def on_joint_changed(index):
            self.joint = self.joint_combo.itemData(index)
            self.setup_motiontype_combo()  # Refresh the motion combo box based on the new joint type

        self.joint_combo.currentIndexChanged.connect(on_joint_changed)

        # Create the line edit for the formula
        formula_edit = QLineEdit(self.dialog)
        formula_edit.setText(self.formula)
        formula_edit.setPlaceholderText(translate("Assembly", "Enter your formula..."))

        # Connect the line edit to update the Formula property
        def on_formula_changed(text):
            self.formula = text

        formula_edit.textChanged.connect(on_formula_changed)

        self.setupHelpSection()

        # Create Ok and Cancel buttons
        button_box = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel, Qt.Horizontal, self.dialog
        )
        button_box.accepted.connect(self.dialog.accept)
        button_box.rejected.connect(self.dialog.reject)

        # Set up the layout of the dialog
        layout = QGridLayout(self.dialog)

        # Add labels and widgets to the layout
        layout.addWidget(QLabel("Joint"), 0, 0)
        layout.addWidget(self.joint_combo, 0, 1)

        layout.addWidget(QLabel("Motion Type"), 1, 0)
        layout.addWidget(self.motion_type_combo, 1, 1)

        layout.addWidget(QLabel("Formula"), 2, 0)
        layout.addWidget(formula_edit, 2, 1)

        # Add the help label above the buttons
        layout.addWidget(self.help_label0, 3, 0, 1, 2)
        layout.addWidget(self.help_label1, 4, 0, 1, 2)
        layout.addWidget(self.help_label2, 5, 0, 1, 2)
        layout.addWidget(self.help_label3, 6, 0, 1, 2)
        layout.addWidget(self.help_label4, 7, 0, 1, 2)
        layout.addWidget(self.help_label5, 8, 0, 1, 2)
        layout.addWidget(self.help_label6, 9, 0, 1, 2)
        layout.addWidget(self.help_label7, 10, 0, 1, 2)
        # Add the help button and button box in the next row
        layout.addWidget(self.help_button, 11, 0)

        layout.addWidget(button_box, 11, 1)

        self.positionDialog()

    def setupHelpSection(self):

        # Create the help QLabels and set them to be initially hidden
        self.help_label0 = QLabel(
            translate(
                "Assembly",
                "In capital are variables that you need to replace with actual values. More details about each example in its tooltip.",
            ),
            self.dialog,
        )
        self.help_label1 = QLabel(translate("Assembly", " - Linear: C + VEL*time"), self.dialog)
        self.help_label2 = QLabel(
            translate("Assembly", " - Quadratic: C + VEL*time + ACC*time^2"), self.dialog
        )
        self.help_label3 = QLabel(
            translate("Assembly", " - Harmonic: C + AMP*sin(VEL*time - PHASE)"), self.dialog
        )
        self.help_label4 = QLabel(
            translate("Assembly", " - Exponential: C*exp(time/TIMEC)"), self.dialog
        )
        self.help_label5 = QLabel(
            translate(
                "Assembly",
                " - Smooth Step: L1 + (L2 - L1)*((1/2) + (1/pi)*arctan(SLOPE*(time - T0)))",
            ),
            self.dialog,
        )
        self.help_label6 = QLabel(
            translate(
                "Assembly",
                " - Smooth Square Impulse: (H/pi)*(arctan(SLOPE*(time - T1)) - arctan(SLOPE*(time - T2)))",
            ),
            self.dialog,
        )
        self.help_label7 = QLabel(
            translate(
                "Assembly",
                " - Smooth Ramp Top Impulse: ((1/pi)*(arctan(1000*(time - T1)) - arctan(1000*(time - T2))))*(((H2 - H1)/(T2 - T1))*(time - T1) + H1)",
            ),
            self.dialog,
        )

        self.help_label1.setToolTip(
            translate(
                "Assembly",
                """C is a constant offset.
VEL is a velocity or slope or gradient of the straight line.""",
            )
        )
        self.help_label2.setToolTip(
            translate(
                "Assembly",
                """C is a constant offset.
VEL is the velocity or slope or gradient of the straight line.
ACC is the acceleration or coefficient of the second order. The function is a parabola.""",
            )
        )
        self.help_label3.setToolTip(
            translate(
                "Assembly",
                """C is a constant offset.
AMP is the amplitude of the sine wave.
VEL is the angular velocity in radians per second.
PHASE is the phase of the sine wave.""",
            )
        )
        self.help_label4.setToolTip(
            translate(
                "Assembly",
                """C is a constant.
TIMEC is the time constant of the exponential function.""",
            )
        )
        self.help_label5.setToolTip(
            translate(
                "Assembly",
                """L1 is step level before time = T0.
L2 is step level after time = T0.
SLOPE defines the steepness of the transition between L1 and L2 about time = T0. Higher values gives sharper cornered steps. SLOPE = 1000 or greater are suitable.""",
            )
        )
        self.help_label6.setToolTip(
            translate(
                "Assembly",
                """H is the height of the impulse.
T1 is the start of the impulse.
T2 is the end of the impulse.
SLOPE defines the steepness of the transition between 0 and H about time = T1 and T2. Higher values gives sharper cornered impulses. SLOPE = 1000 or greater are suitable.""",
            )
        )
        self.help_label7.setToolTip(
            translate(
                "Assembly",
                """This is similar to the square impulse but the top has a sloping ramp. It is good for building a smooth piecewise linear function by adding a series of these.
T1 is the start of the impulse.
T2 is the end of the impulse.
H1 is the height at T1 at the beginning of the ramp.
H2 is the height at T2 at the end of the ramp.
SLOPE defines the steepness of the transition between 0 and H1 and H2 to 0 about time = T1 and T2 respectively. Higher values gives sharper cornered impulses. SLOPE = 1000 or greater are suitable.""",
            )
        )

        self.help_label0.setWordWrap(True)
        self.help_label1.setWordWrap(True)
        self.help_label2.setWordWrap(True)
        self.help_label3.setWordWrap(True)
        self.help_label4.setWordWrap(True)
        self.help_label5.setWordWrap(True)
        self.help_label6.setWordWrap(True)
        self.help_label7.setWordWrap(True)

        width = 1000
        self.help_label0.setFixedWidth(width)
        self.help_label1.setFixedWidth(width)
        self.help_label2.setFixedWidth(width)
        self.help_label3.setFixedWidth(width)
        self.help_label4.setFixedWidth(width)
        self.help_label5.setFixedWidth(width)
        self.help_label6.setFixedWidth(width)
        self.help_label7.setFixedWidth(width)

        self.help_label0.setVisible(False)
        self.help_label1.setVisible(False)
        self.help_label2.setVisible(False)
        self.help_label3.setVisible(False)
        self.help_label4.setVisible(False)
        self.help_label5.setVisible(False)
        self.help_label6.setVisible(False)
        self.help_label7.setVisible(False)

        self.help_label1.setTextInteractionFlags(Qt.TextSelectableByMouse)
        self.help_label2.setTextInteractionFlags(Qt.TextSelectableByMouse)
        self.help_label3.setTextInteractionFlags(Qt.TextSelectableByMouse)
        self.help_label4.setTextInteractionFlags(Qt.TextSelectableByMouse)
        self.help_label5.setTextInteractionFlags(Qt.TextSelectableByMouse)
        self.help_label6.setTextInteractionFlags(Qt.TextSelectableByMouse)
        self.help_label7.setTextInteractionFlags(Qt.TextSelectableByMouse)
        # Create the Help button
        self.help_button = QPushButton(translate("Assembly", "Help"), self.dialog)

        # Slot to toggle help visibility and button text
        def toggle_help():
            show = not self.help_label1.isVisible()
            self.help_label0.setVisible(show)
            self.help_label1.setVisible(show)
            self.help_label2.setVisible(show)
            self.help_label3.setVisible(show)
            self.help_label4.setVisible(show)
            self.help_label5.setVisible(show)
            self.help_label6.setVisible(show)
            self.help_label7.setVisible(show)

            if show:
                self.help_button.setText(translate("Assembly", "Hide help"))
            else:
                self.help_button.setText(translate("Assembly", "Help"))

            self.positionDialog()

        self.help_button.clicked.connect(toggle_help)

    def positionDialog(self):
        self.dialog.adjustSize()

        # Get the screen where the mouse is located
        screen = QGuiApplication.screenAt(self.initialPos)
        screen_geometry = (
            screen.availableGeometry()
            if screen
            else QApplication.primaryScreen().availableGeometry()
        )

        # Calculate the position of the dialog to ensure it stays within the screen
        dialog_position = self.initialPos

        # Adjust position to keep the dialog within the screen bounds
        if dialog_position.x() + self.dialog.width() > screen_geometry.right():
            dialog_position.setX(screen_geometry.right() - self.dialog.width())
        if dialog_position.y() + self.dialog.height() > screen_geometry.bottom():
            dialog_position.setY(screen_geometry.bottom() - self.dialog.height())

        # Ensure the dialog does not go above or to the left of the screen
        if dialog_position.x() < screen_geometry.left():
            dialog_position.setX(screen_geometry.left())
        if dialog_position.y() < screen_geometry.top():
            dialog_position.setY(screen_geometry.top())

        # Move the dialog to the final position
        self.dialog.move(dialog_position)

    def setup_joint_combo(self):
        # Function to set up the joint combo box based on the selected motion type

        self.joint_combo.clear()  # Clear existing items

        jointTypes = ["Revolute", "Slider", "Cylindrical"]

        joints = UtilsAssembly.getJointsOfType(self.assembly, jointTypes)

        # Add joints to the combo box with labels and icons
        for joint in joints:
            joint_label = joint.Label
            joint_icon = QIcon(joint.ViewObject.Icon)
            self.joint_combo.addItem(joint_icon, joint_label, userData=joint)

        # Set the current value based on the object's Joint property
        if self.joint in joints:
            self.joint_combo.setCurrentText(self.joint.Label)
        elif len(joints) > 0:
            self.joint = joints[0]

    def setup_motiontype_combo(self):
        self.motion_type_combo.clear()  # Clear existing items

        if self.joint is None:
            return

        if self.joint.JointType == "Revolute":
            types = ["Angular"]
        elif self.joint.JointType == "Slider":
            types = ["Linear"]
        else:
            types = ["Angular", "Linear"]

        self.motion_type_combo.addItems(types)

        # Set current value based on the object's MotionType
        if self.motionType in types:
            self.motion_type_combo.setCurrentText(self.motionType)
        else:
            # self.motionType is no longer available, so we reset it to first entry
            self.motionType = types[0]

    def exec_(self):
        return self.dialog.exec()


######### Create Simulation Task ###########
class TaskAssemblyCreateSimulation(QtCore.QObject):
    def __init__(self, simFeaturePy=None):
        super().__init__()
        Gui.Selection.clearSelection()

        self.assembly = UtilsAssembly.activeAssembly()

        self.initialPlcs = UtilsAssembly.saveAssemblyPartsPlacements(self.assembly)

        self.doc = self.assembly.Document
        self.gui_doc = Gui.getDocument(self.doc)

        self.view = self.gui_doc.activeView()

        if not self.assembly or not self.view or not self.doc:
            return

        self.runKinematicsTimer = QtCore.QTimer()
        self.runKinematicsTimer.setSingleShot(True)
        self.runKinematicsTimer.timeout.connect(self.displayLastFrame)

        self.animationTimer = QtCore.QTimer()
        self.animationTimer.setInterval(50)  # ms
        self.animationTimer.timeout.connect(self.playAnimation)

        self.form = Gui.PySideUic.loadUi(":/panels/TaskAssemblyCreateSimulation.ui")
        self.form.motionList.installEventFilter(self)
        self.setSpinboxPrecision(self.form.TimeStartSpinBox, 9)
        self.setSpinboxPrecision(self.form.TimeEndSpinBox, 9)
        self.setSpinboxPrecision(self.form.TimeStepOutputSpinBox, 9)
        self.setSpinboxPrecision(self.form.GlobalErrorToleranceSpinBox, 9, App.Units.Length)
        self.form.motionList.itemDoubleClicked.connect(self.onItemDoubleClicked)
        self.form.TimeStartSpinBox.valueChanged.connect(self.onTimeStartChanged)
        self.form.TimeEndSpinBox.valueChanged.connect(self.onTimeEndChanged)
        self.form.TimeStepOutputSpinBox.valueChanged.connect(self.onTimeStepOutputChanged)
        self.form.GlobalErrorToleranceSpinBox.valueChanged.connect(
            self.onGlobalErrorToleranceChanged
        )
        self.form.RunKinematicsButton.clicked.connect(self.runKinematics)
        self.form.frameSlider.valueChanged.connect(self.onFrameChanged)
        self.form.FramesPerSecondSpinBox.valueChanged.connect(self.onFramesPerSecondChanged)
        self.form.PlayBackwardButton.clicked.connect(self.animationTimerStartBackward)
        self.form.PlayForwardButton.clicked.connect(self.animationTimerStartForward)
        self.form.StepBackwardButton.clicked.connect(self.stepBackward)
        self.form.StepForwardButton.clicked.connect(self.stepForward)
        self.form.StopButton.clicked.connect(self.stopAnimation)
        self.form.AddButton.clicked.connect(self.addMotionClicked)
        self.form.RemoveButton.clicked.connect(self.deleteSelectedMotions)
        self.form.groupBox_player.hide()

        if simFeaturePy:
            self.simFeaturePy = simFeaturePy
            App.setActiveTransaction("Edit " + simFeaturePy.Label + " Simulation")
            self.onMotionsChanged()
        else:
            App.setActiveTransaction("Create Simulation")
            self.createSimulationObject()

        self.setUiInitialValues()

        self.simFeaturePy.Proxy.setMotionsChangedCallback(self.onMotionsChanged)

        self.currentFrm = 1
        self.startFrm = 1
        self.endFrm = 100
        self.fps = 30
        self.deltaTime = 1.0 / self.fps
        self.startTime = time.time()
        self.index = 0

    def setUiInitialValues(self):
        self.form.TimeStartSpinBox.setProperty("rawValue", self.simFeaturePy.aTimeStart.Value)
        self.form.TimeEndSpinBox.setProperty("rawValue", self.simFeaturePy.bTimeEnd.Value)
        self.form.TimeStepOutputSpinBox.setProperty(
            "rawValue", self.simFeaturePy.cTimeStepOutput.Value
        )
        self.form.GlobalErrorToleranceSpinBox.setProperty(
            "rawValue", self.simFeaturePy.fGlobalErrorTolerance
        )
        self.setFrameValue(0)
        self.form.FramesPerSecondSpinBox.setValue(self.simFeaturePy.jFramesPerSecond)

    def setSpinboxPrecision(self, spinbox, precision, unit=App.Units.TimeSpan):
        q = App.Units.Quantity()
        q.Unit = unit
        q.Format = {"Precision": precision}
        spinbox.setProperty("value", q)

    def accept(self):
        self.deactivate()
        UtilsAssembly.restoreAssemblyPartsPlacements(self.assembly, self.initialPlcs)
        App.closeActiveTransaction()
        return True

    def reject(self):
        self.deactivate()
        App.closeActiveTransaction(True)
        return True

    def deactivate(self):
        self.animationTimer.stop()
        self.simFeaturePy.Proxy.setMotionsChangedCallback(None)
        if Gui.Control.activeDialog():
            Gui.Control.closeDialog()

    def onTimeStartChanged(self, quantity):
        self.simFeaturePy.aTimeStart = self.form.TimeStartSpinBox.property("rawValue")

    def onTimeEndChanged(self, quantity):
        self.simFeaturePy.bTimeEnd = self.form.TimeEndSpinBox.property("rawValue")

    def onTimeStepOutputChanged(self, quantity):
        self.simFeaturePy.cTimeStepOutput = self.form.TimeStepOutputSpinBox.property("rawValue")

    def onGlobalErrorToleranceChanged(self, quantity):
        self.simFeaturePy.fGlobalErrorTolerance = self.form.GlobalErrorToleranceSpinBox.property(
            "rawValue"
        )

    def onItemDoubleClicked(self, item):
        row = self.form.motionList.row(item)
        if row < len(self.simFeaturePy.Group):
            motion = self.simFeaturePy.Group[row]
            motion.ViewObject.Proxy.openEditDialog()
            self.onMotionsChanged()

    def createSimulationObject(self):
        sim_group = UtilsAssembly.getSimulationGroup(self.assembly)
        self.simFeaturePy = sim_group.newObject("App::FeaturePython", "Simulation")
        Simulation(self.simFeaturePy)
        ViewProviderSimulation(self.simFeaturePy.ViewObject)

    def createMotionObject(self, motionType, joint, formula):
        motion = self.assembly.newObject("App::FeaturePython", "Motion")
        Motion(motion, motionType, joint, formula)
        ViewProviderMotion(motion.ViewObject)

        listOfMotions = self.simFeaturePy.Group
        listOfMotions.append(motion)
        self.simFeaturePy.Group = listOfMotions

    def onMotionsChanged(self):
        self.form.motionList.clear()
        for motion in self.simFeaturePy.Group:
            self.form.motionList.addItem(motion.Label)

    def runKinematics(self):
        self.assembly.generateSimulation(self.simFeaturePy)
        nFrms = self.assembly.numberOfFrames()
        self.form.frameSlider.setMaximum(nFrms - 1)
        self.setFrameValue(nFrms - 1)
        self.form.groupBox_player.show()

    def onFrameChanged(self, val):
        self.assembly.updateForFrame(val)
        self.form.FrameLabel.setText(translate("Assembly", "Frame" + " " + str(val)))
        time = float(val * self.simFeaturePy.cTimeStepOutput)
        self.form.FrameTimeLabel.setText(f"{time:.2f} s")

    def onFramesPerSecondChanged(self):
        self.simFeaturePy.jFramesPerSecond = self.form.FramesPerSecondSpinBox.value()

    def playBackward(self):
        pass

    def animationTimerStartForward(self):
        self.direction = 1
        self.animationTimerStart()

    def animationTimerStartBackward(self):
        self.direction = -1
        self.animationTimerStart()

    def animationTimerStart(self):
        self.animationTimer.stop()
        self.currentFrm = self.form.frameSlider.value()
        self.startFrm = 1
        self.endFrm = self.form.frameSlider.maximum()
        if self.startFrm >= self.endFrm:
            return

        self.fps = self.simFeaturePy.jFramesPerSecond
        self.deltaTime = 1.0 / self.fps
        self.startTime = time.time()
        self.index = self.currentFrm
        self.animationTimer.setInterval(self.deltaTime * 1000)  # ms
        self.animationTimer.start()

    def playAnimation(self):
        range_ = self.endFrm - self.startFrm
        offset = self.currentFrm - self.startFrm
        count = int((time.time() - self.startTime) / self.deltaTime)
        self.index = ((self.direction * count + offset) % range_) + self.startFrm
        self.setFrameValue(self.index)

    def displayLastFrame(self):
        nFrms = self.assembly.numberOfFrames()
        self.setFrameValue(nFrms - 1)

    def stepBackward(self):
        self.animationTimer.stop()

        nextFrm = self.form.frameSlider.value() - 1
        if nextFrm < 1:
            nextFrm = self.form.frameSlider.maximum()  # wraparound
        self.setFrameValue(nextFrm)

    def stepForward(self):
        self.animationTimer.stop()

        nextFrm = self.form.frameSlider.value() + 1
        if nextFrm > self.form.frameSlider.maximum():
            nextFrm = 1  # wraparound
        self.setFrameValue(nextFrm)

    def setFrameValue(self, val):
        if val < 1:
            val = 1
        if val > self.form.frameSlider.maximum():
            val = self.form.frameSlider.maximum()

        self.form.frameSlider.setValue(val)

    def stopAnimation(self):
        self.animationTimer.stop()

    def addMotionClicked(self):
        dialog = MotionEditDialog(self.assembly)
        if dialog.exec_():
            self.createMotionObject(dialog.motionType, dialog.joint, dialog.formula)

    # Taskbox keyboard event handler
    def eventFilter(self, watched, event):
        if self.form is not None and watched == self.form.motionList:
            if event.type() == QtCore.QEvent.ShortcutOverride:
                if event.key() == QtCore.Qt.Key_Delete:
                    event.accept()
                    return True  # Indicate that the event has been handled
                return False

            elif event.type() == QtCore.QEvent.KeyPress:
                if event.key() == QtCore.Qt.Key_Delete:
                    self.deleteSelectedMotions()
                    return True  # Consume the event

        return super().eventFilter(watched, event)

    def deleteSelectedMotions(self):
        selected_indexes = self.form.motionList.selectedIndexes()
        sorted_indexes = sorted(selected_indexes, key=lambda x: x.row(), reverse=True)
        for index in sorted_indexes:
            row = index.row()
            if row < len(self.simFeaturePy.Group):
                motion = self.simFeaturePy.Group[row]
                # First remove the link from the viewObj
                self.simFeaturePy.Group.remove(motion)
                # Delete the object
                motion.Document.removeObject(motion.Name)


if App.GuiUp:
    Gui.addCommand("Assembly_CreateSimulation", CommandCreateSimulation())
